import os
import os.path
import git
from git.repo.fun import is_git_dir
import logging
from six.moves.urllib.parse import urlparse
from unfurl.util import UnfurlError

logger = logging.getLogger("unfurl")


def normalizeGitUrl(url):
    if url.startswith("git-local://"):  # truncate url after commit digest
        return "git-local://" + urlparse(url).netloc.partition(":")[0]
    return url


def isURLorGitPath(url):
    if "://" in url and not url.startswith("file:"):
        return True
    if "@" in url:
        return True
    candidate, sep, frag = url.partition("#")
    if candidate.rstrip("/").endswith(".git"):
        return True
    return False


def splitGitUrl(url):
    """
  Returns (repoURL, filePath, revision)
  RepoURL will be an empty string if it isn't a path to a git repo
  """
    parts = urlparse(url)
    if parts.scheme == "git-local":
        return parts.scheme + "://" + parts.netloc, parts.path[1:], parts.fragment

    if parts.fragment:
        # treat fragment as a git revision spec; see https://git-scm.com/docs/gitrevisions
        # or https://docs.docker.com/engine/reference/commandline/build/#git-repositories
        # just support <ref>:<path> for now
        # e.g. myrepo.git#mybranch, myrepo.git#pull/42/head, myrepo.git#:myfolder, myrepo.git#master:myfolder
        revision, sep, path = parts.fragment.partition(":")
        giturl, sep, frag = url.partition("#")
        return giturl, path, revision
    return url, "", ""


class _ProgressPrinter(git.RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=""):
        # print update to stdout but only if logging is INFO or more verbose
        if message and logger.getEffectiveLevel() <= logging.INFO:
            print("fetching from %s, received: %s " % (self.gitUrl, message))


class Repo(object):
    @staticmethod
    def findContainingRepo(rootDir, gitDir=".git"):
        """
    Walk parents looking for a git repository.
    """
        current = os.path.abspath(rootDir)
        while current and current != os.sep:
            if is_git_dir(os.path.join(current, gitDir)):
                return GitRepo(git.Repo(current))
            current = os.path.dirname(current)
        return None

    @staticmethod
    def findGitWorkingDirs(rootDir, gitDir=".git"):
        workingDirs = {}
        for root, dirs, files in os.walk(rootDir):
            if Repo.updateGitWorkingDirs(workingDirs, root, dirs, gitDir):
                del dirs[:]  # don't visit sub directories
        return workingDirs

    @staticmethod
    def updateGitWorkingDirs(workingDirs, root, dirs, gitDir=".git"):
        if gitDir in dirs and is_git_dir(os.path.join(root, gitDir)):
            assert os.path.isdir(root), root
            repo = GitRepo(git.Repo(root))
            key = os.path.abspath(root)
            workingDirs[key] = (repo.url, repo)
            return key
        return None

    @staticmethod
    def ignoreDir(dir):
        parent = Repo.findContainingRepo(os.path.dirname(dir))
        if parent:
            path = parent.findRepoPath(dir)
            if path:  # can be None if dir is already ignored
                parent.addToLocalGitIgnore(path)
                return path
        return None

    def findRepoPath(self, path):
        localPath = self.findPath(path)[0]
        if localPath is not None and not self.isPathExcluded(localPath):
            return localPath
        return None

    def isPathExcluded(self, localPath):
        return False

    def findPath(self, path, importLoader=None):
        base = self.workingDir
        if not base:  # XXX support bare repos
            return None, None, None
        repoRoot = os.path.abspath(base)
        abspath = os.path.abspath(path).rstrip("/")
        if repoRoot in abspath:
            # XXX find pinned
            # if importLoader:
            #   revision = importLoader.getRevision(self)
            # else:
            if True:
                revision = self.revision
            bare = not self.workingDir or revision != self.revision
            return abspath[len(repoRoot) + 1 :], revision, bare
        return None, None, None

    @classmethod
    def createWorkingDir(cls, gitUrl, localRepoPath, revision=None):
        branch = revision or "master"
        logger.info("Fetching %s (%s) to %s", gitUrl, branch, localRepoPath)
        progress = _ProgressPrinter()
        progress.gitUrl = gitUrl
        try:
            repo = git.Repo.clone_from(gitUrl, localRepoPath, progress, branch=branch)
        except git.exc.GitCommandError as err:
            raise UnfurlError(
                "couldn't create working dir, clone failed: " + err._cmdline
            )
        Repo.ignoreDir(localRepoPath)
        return GitRepo(repo)


class GitRepo(Repo):
    def __init__(self, gitrepo):
        self.repo = gitrepo
        self.url = self.workingDir or gitrepo.git_dir
        if gitrepo.remotes:
            # note: these might not look like absolute urls, e.g. git@github.com:onecommons/unfurl.git
            try:
                remote = gitrepo.remotes["origin"]
            except:
                remote = gitrepo.remotes[0]
            self.url = remote.url

    @property
    def workingDir(self):
        dir = self.repo.working_tree_dir
        if not dir or dir[-1] == "/":
            return dir
        else:
            return dir + "/"

    @property
    def revision(self):
        return self.repo.head.commit.hexsha

    def resolveRevSpec(self, revision):
        try:
            return self.repo.commit(revision).hexsha
        except:
            return None

    def getUrlWithPath(self, path):
        if isURLorGitPath(self.url):
            if os.path.isabs(path):
                # get path relative to repository's root
                path = os.path.relpath(path, self.workingDir)
            return self.url + "#:" + path
        else:
            return self.getGitLocalUrl(path)

    def findExcludedDirs(self, root):
        root = os.path.relpath(root, self.workingDir)
        status, stdout, stderr = self.runCmd(
            [
                "ls-files",
                "--exclude-standard",
                "-o",
                "-i",
                "--full-name",
                "--directory",
                root,
            ]
        )
        for file in stdout.splitlines():
            path = os.path.join(self.workingDir, file)
            yield path

    def isPathExcluded(self, localPath):
        # XXX cache and test
        # excluded = list(self.findExcludedDirs(self.workingDir))
        # success error code means it's ignored
        return not self.runCmd(["check-ignore", "-q", localPath])[0]

    def runCmd(self, args, **kw):
        """
    :return:
      tuple(int(status), str(stdout), str(stderr))
    """
        gitcmd = self.repo.git
        call = [gitcmd.GIT_PYTHON_GIT_EXECUTABLE]
        # add persistent git options
        call.extend(gitcmd._persistent_git_options)
        call.extend(list(args))

        # note: sets cwd to working_dir
        return gitcmd.execute(
            call, with_exceptions=False, with_extended_output=True, **kw
        )

    def addToLocalGitIgnore(self, rule):
        with open(os.path.join(self.repo.git_dir, "info", "exclude"), "a") as f:
            f.write("\n" + rule)

    def show(self, path, commitId):
        if self.workingDir and os.path.isabs(path):
            path = os.path.abspath(path)[len(self.workingDir) :]
        # XXX this won't work if path is in a submodule
        # if in path startswith a submodule: git log -1 -p [commitid] --  [submodule]
        # submoduleCommit = re."\+Subproject commit (.+)".group(1)
        # return self.repo.submodules[submodule].git.show(submoduleCommit+':'+path[len(submodule)+1:])
        return self.repo.git.show(commitId + ":" + path)

    def checkout(self, revision=""):
        # if revision isn't specified and repo is not pinned:
        #  save the ref of current head
        self.repo.git.checkout(revision)
        logger.info(
            "checking out '%s' at %s to %s",
            self.url,
            revision or "HEAD",
            self.workingDir,
        )
        return self.workingDir

    def getInitialRevision(self):
        firstCommit = next(self.repo.iter_commits("HEAD", max_parents=0))
        return firstCommit.hexsha

    def commitFiles(self, files, msg):
        # note: this will also commit existing changes in the index
        index = self.repo.index
        index.add([os.path.abspath(f) for f in files])
        return index.commit(msg)

    def isDirty(self, untracked_files=False):
        # diff = self.repo.git.diff()  # "--abbrev=40", "--full-index", "--raw")
        return self.repo.is_dirty(untracked_files=untracked_files)

    def clone(self, newPath):
        cloned = self.repo.clone(os.path.abspath(newPath))
        Repo.ignoreDir(newPath)
        return GitRepo(cloned)

    def getGitLocalUrl(self, path, name=""):
        if os.path.isabs(path):
            # get path relative to repository's root
            path = os.path.relpath(path, self.workingDir)
        return "git-local://%s:%s/%s" % (self.getInitialRevision(), name, path)

    # XXX: def getDependentRepos()
    # XXX: def canManage()

    # def canMakeClean(self):
    #     for repo in self.getDependentRepos():
    #         if not repo.canMakeClean():
    #             return False
    #         elif repo.isDirty() and not self.canManage(repo):
    #             return False
    #     return True
    #
    # def _commitAll(self, parent=None):
    #     committed = []
    #     for repo in self.getDependentRepos():
    #         if repo.isDirty():
    #             assert self.canManage(repo)
    #             repo._commitAll(self)
    #             committed.append(repo)
    #     self.updateChildCommits(committed)
    #     self._commit()
    #
    # def getDirtyDependents(self):
    #     for repo in self.getDependentRepos():
    #         if repo.isDirty():
    #             yield repo

    # XXX unused.. currently yamlmanifest.commitJob() calls commitFiles()
    # def commit(self):
    #     # before run referenced dirty repos should be committed?
    #     # at the very least the state of any declared repo should be saved
    #     # otherwise two different runs of the same commit could pull different versions
    #     # this is true for the spec repos also -- save in spec's manifest-template?
    #     repo = self.repo
    #     repo.index.add("*")
    #     # commit the manifest first so we can get a commit ref for the changerecord
    #     commit = repo.git.commit("")
    #     changeFiles = self.manifest.saveChanges(commit.hexsha)
    #     repo.index.add(changeFiles)
    #     repo.git.commit("")


# class SimpleRepo(Repo):
#   def __init__(self, lastCommitId):
#     self.lastCommitId = int(lastCommitId or 0)
#
#   def checkout(self, commitid, useCurrent):
#     if useCurrent and commitid == self.lastCommitId:
#       return self.workingDir
#     return './revisions/{commitid}/files'
#
#   def commit(self):
#     self.lastCommitId += 1
#     # copy current workingDir to /revisions/{commitid}/files
#     return self.lastCommitId


class RevisionManager(object):
    def __init__(self, manifest, localEnv=None):
        self.manifest = manifest
        self.revisions = None
        self.localEnv = localEnv

    def getRevision(self, change):
        if self.revisions is None:
            self.revisions = {self.manifest.specDigest: self.manifest}
        digest = change["specDigest"]
        commitid = change["startCommit"]
        if digest in self.revisions:
            return self.revisions[digest]
        else:
            from .manifest import SnapShotManifest

            manifest = SnapShotManifest(self.manifest, commitid)
            self.revisions[digest] = manifest
            return manifest



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import InvalidTypeAdditionalRequirementsError
from toscaparser.utils.gettextutils import _
import toscaparser.utils.validateutils as validateutils


log = logging.getLogger('tosca')


class PortSpec(dict):
    '''Parent class for tosca.datatypes.network.PortSpec type.'''

    SHORTNAME = 'PortSpec'
    TYPE_URI = 'tosca.datatypes.network.' + SHORTNAME

    PROPERTY_NAMES = (
        PROTOCOL, SOURCE, SOURCE_RANGE,
        TARGET, TARGET_RANGE
    ) = (
        'protocol', 'source', 'source_range',
        'target', 'target_range'
    )

    def __str__(self):
        """
        Translate a `tosca:PortSpec` into a string like "source-range:target-range/udp"
        or "source:target" or just "source" if target is missing or the same.
        """
        if "source_range" in self:
            source = "{:d}-{:d}".format(*self["source_range"])
        else:
            source = "{}".format(self.get("source", ""))
        if "target_range" in self:
            target = "{:d}-{:d}".format(*self["target_range"])
        else:
            target = "{}".format(self.get("target", ""))
        if source == target or not target:
            portSpec = str(source)
        else:
            portSpec = "{}:{}".format(target, source)
        protocol = self.get("protocol", "tcp")
        if protocol != "tcp":
            portSpec += "/" + protocol
        return portSpec

    # The following additional requirements MUST be tested:
    # 1) A valid PortSpec MUST have at least one of the following properties:
    #   target, target_range, source or source_range.
    # 2) A valid PortSpec MUST have a value for the source property that
    #    is within the numeric range specified by the property source_range
    #    when source_range is specified.
    # 3) A valid PortSpec MUST have a value for the target property that is
    #    within the numeric range specified by the property target_range
    #    when target_range is specified.
    def validate(self):
        from toscaparser.dataentity import DataEntity
        # TODO(TBD) bug 1567063, validate source & target as PortDef type
        # not just as integers
        try:
            source = self.get(PortSpec.SOURCE)
            source_range = self.get(PortSpec.SOURCE_RANGE)
            target = self.get(PortSpec.TARGET)
            target_range = self.get(PortSpec.TARGET_RANGE)

            # verify one of the specified values is set
            if source is None and source_range is None and \
                    target is None and target_range is None:
                ExceptionCollector.appendException(
                    InvalidTypeAdditionalRequirementsError(
                        type=PortSpec.TYPE_URI))
            # Validate source value is in specified range
            if source:
                if source_range:
                    validateutils.validate_value_in_range(source, source_range,
                                                      PortSpec.SOURCE)
                else:
                    portdef = DataEntity('PortDef', source, None, PortSpec.SOURCE)
                    portdef.validate()
            # Validate target value is in specified range
            if target:
                if target_range:
                    validateutils.validate_value_in_range(target, target_range,
                                                      PortSpec.TARGET)
                else:
                    portdef = DataEntity('PortDef', target, None, PortSpec.TARGET)
                    portdef.validate()
        except Exception:
            msg = _('"%(value)s" do not meet requirements '
                    'for type "%(type)s".') \
                % {'value': self, 'type': PortSpec.SHORTNAME}
            ExceptionCollector.appendException(
                ValueError(msg))



import collections
import os.path
import hashlib
import json
from ruamel.yaml.comments import CommentedMap
from .tosca import ToscaSpec, TOSCA_VERSION, Artifact

from .support import ResourceChanges, AttributeManager, Status, Priority, NodeState
from .runtime import (
    OperationalInstance,
    NodeInstance,
    CapabilityInstance,
    RelationshipInstance,
)
from .util import UnfurlError, toEnum, sensitive_str, getBaseDir
from .repo import RevisionManager, splitGitUrl, isURLorGitPath
from .yamlloader import YamlConfig, loadYamlFromArtifact, yaml
from .job import ConfigChange

# from .configurator import Dependency
import logging

logger = logging.getLogger("unfurl")

_basepath = os.path.abspath(os.path.dirname(__file__))


class Manifest(AttributeManager):
    """
  Loads a model from dictionary representing the manifest
  """

    rootResource = None

    def __init__(self, path, localEnv=None):
        super(Manifest, self).__init__(yaml)
        self.localEnv = localEnv
        self.path = path
        self.repo = self._findRepo()
        self.currentCommitId = self.repo and self.repo.revision
        self.revisions = RevisionManager(self)
        self.changeSets = None
        self.tosca = None
        self.specDigest = None

    def _setSpec(self, spec):
        self._setRepositories(spec)
        self.tosca = self.loadSpec(spec, self.path)
        self.specDigest = self.getSpecDigest(spec)

    def _findRepo(self):
        # check if this path exists in the repo
        repo = self.localEnv and self.localEnv.instanceRepo
        if repo:
            path = repo.findPath(self.path)[0]
            if path and (path, 0) in repo.repo.index.entries:
                return repo
        return None

    def loadSpec(self, spec, path):
        if "service_template" in spec:
            toscaDef = spec["service_template"]
        elif "tosca" in spec:  # backward compat
            toscaDef = spec["tosca"]
        else:
            toscaDef = {}

        # make sure this is present
        toscaDef["tosca_definitions_version"] = TOSCA_VERSION

        if not isinstance(toscaDef, CommentedMap):
            toscaDef = CommentedMap(toscaDef.items())
        # hack so we can make the manifest accessible to the yamlloader:
        toscaDef.manifest = self
        if getattr(toscaDef, "baseDir", None) and (
            not path or toscaDef.baseDir != os.path.dirname(path)
        ):
            # note: we only recorded the baseDir not the name of the included file
            path = toscaDef.baseDir
        return ToscaSpec(toscaDef, spec.get("inputs"), spec, path)

    def getSpecDigest(self, spec):
        m = hashlib.sha1()  # use same digest function as git
        t = self.tosca.template
        for tpl in [spec, t.topology_template.custom_defs, t.nested_tosca_tpls]:
            m.update(json.dumps(tpl, sort_keys=True).encode("utf-8"))
        return m.hexdigest()

    def _ready(self, rootResource):
        self.rootResource = rootResource
        if rootResource:
            rootResource.attributeManager = self

    def getRootResource(self):
        return self.rootResource

    def getBaseDir(self):
        return "."

    @property
    def repositories(self):
        return self.tosca.template.repositories

    @property
    def loader(self):
        return (
            self.rootResource
            and self.rootResource.templar
            and self.rootResource.templar._loader
        )

    def saveJob(self, job):
        pass

    def loadTemplate(self, name, lastChange=None):
        if lastChange:
            try:
                return self.revisions.getRevision(lastChange).tosca.getTemplate(name)
            except:
                return None
        else:
            return self.tosca.getTemplate(name)

    #  load instances
    #    create a resource with the given template
    #  or generate a template setting interface with the referenced implementations

    @staticmethod
    def loadStatus(status, instance=None):
        if not instance:
            instance = OperationalInstance()
        if not status:
            return instance

        instance._priority = toEnum(Priority, status.get("priority"))
        instance._lastStateChange = status.get("lastStateChange")
        instance._lastConfigChange = status.get("lastConfigChange")

        readyState = status.get("readyState")
        if not isinstance(readyState, collections.Mapping):
            instance._localStatus = toEnum(Status, readyState)
        else:
            instance._localStatus = toEnum(Status, readyState.get("local"))
            instance._state = toEnum(NodeState, readyState.get("state"))

        return instance

    @staticmethod
    def loadResourceChanges(changes):
        resourceChanges = ResourceChanges()
        if changes:
            for k, change in changes.items():
                status = change.pop(".status", None)
                resourceChanges[k] = [
                    None if status is None else Manifest.loadStatus(status).localStatus,
                    change.pop(".added", None),
                    change,
                ]
        return resourceChanges

    # XXX unused
    # def loadConfigChange(self, changeId):
    #     """
    # Reconstruct the Configuration that was applied in the past
    # """
    #     changeSet = self.changeSets.get(changeId)
    #     if not changeSet:
    #         raise UnfurlError("can not find changeset for changeid %s" % changeId)
    #
    #     configChange = ConfigChange()
    #     Manifest.loadStatus(changeSet, configChange)
    #     configChange.changeId = changeSet.get("changeId", 0)
    #     configChange.parentId = changeSet.get("parentId")
    #
    #     configChange.inputs = changeSet.get("inputs")
    #
    #     configChange.dependencies = {}
    #     for val in changeSet.get("dependencies", []):
    #         key = val.get("name") or val["ref"]
    #         assert key not in configChange.dependencies
    #         configChange.dependencies[key] = Dependency(
    #             val["ref"],
    #             val.get("expected"),
    #             val.get("schema"),
    #             val.get("name"),
    #             val.get("required"),
    #             val.get("wantList", False),
    #         )
    #
    #     if "changes" in changeSet:
    #         configChange.resourceChanges = self.loadResourceChanges(
    #             changeSet["changes"]
    #         )
    #
    #     configChange.result = changeSet.get("result")
    #     configChange.messages = changeSet.get("messages", [])
    #
    #     # XXX
    #     # ('action', ''),
    #     # ('target', ''), # nodeinstance key
    #     # implementationType: configurator resource | artifact | configurator class
    #     # implementation: repo:key#commitid | className:version
    #     return configChange

    # XXX this is only used by commented out Taskview.createConfigurationSpec()
    # def loadConfigSpec(self, configName, spec):
    #  # XXX need to update configurationSpec in json schema:
    #  #"workflow": { "type": "string" },
    #  #"inputs": { "$ref": "#/definitions/attributes", "default": {} },
    #  #"preconditions": { "$ref": "#/definitions/schema", "default": {} }
    #     return ConfigurationSpec(
    #         configName,
    #         spec["operation"],
    #         spec["className"],
    #         spec.get("majorVersion"),
    #         spec.get("minorVersion", ""),
    #         workflow=spec.get("workflow", Defaults.workflow),
    #         inputs=spec.get("inputs"),
    #         inputSchema=spec.get("inputSchema"),
    #         preConditions=spec.get("preConditions"),
    #         postConditions=spec.get("postConditions"),
    #     )

    def createNodeInstance(self, rname, resourceSpec, parent=None):
        # if parent property is set it overrides the parent argument
        pname = resourceSpec.get("parent")
        if pname:
            parent = self.getRootResource().findResource(pname)
            if parent is None:
                raise UnfurlError("can not find parent instance %s" % pname)

        resource = self._createEntityInstance(NodeInstance, rname, resourceSpec, parent)

        resource._capabilities = []
        for key, val in resourceSpec.get("capabilities", {}).items():
            self._createEntityInstance(CapabilityInstance, key, val, resource)

        resource._requirements = []
        for key, val in resourceSpec.get("requirements", {}).items():
            # parent will be the capability, should have already been created
            capabilityId = val.get("capability")
            if not capabilityId:
                raise UnfurlError("requirement is missing capability %s" % key)
            capability = capabilityId and self.getRootResource().query(capabilityId)
            if not capability or not isinstance(capability, CapabilityInstance):
                raise UnfurlError("can not find capability %s" % capabilityId)
            if capability._relationships is None:
                capability._relationships = []
            requirement = self._createEntityInstance(
                RelationshipInstance, key, val, capability
            )
            requirement.source = resource
            resource.requirements.append(requirement)

        for key, val in resourceSpec.get("instances", {}).items():
            self.createNodeInstance(key, val, resource)

        return resource

    def _getLastChange(self, operational):
        if not operational.lastConfigChange:
            return None
        if not self.changeSets:  # XXX load changesets if None
            return None
        jobId = ConfigChange.getJobId(operational.lastConfigChange)
        return self.changeSets.get(jobId)

    def _createEntityInstance(self, ctor, name, status, parent):
        operational = self.loadStatus(status)
        templateName = status.get("template", name)

        imported = None
        importName = status.get("imported")
        if importName is not None:
            imported = self.imports.findImport(importName)
            if not imported:
                raise UnfurlError("missing import %s" % importName)
            if imported.shadow:
                raise UnfurlError("already imported %s" % importName)
            template = imported.template
        else:
            template = self.loadTemplate(templateName)

        if template is None:
            # not defined in the current model any more, try to retrieve the old version
            if operational.lastConfigChange:
                changerecord = self._getLastChange(operational)
                template = self.loadTemplate(templateName, changerecord)
        if template is None:
            raise UnfurlError("missing resource template %s" % templateName)
        # logger.debug("creating instance for template %s: %s", templateName, template)

        # omit keys that match <<REDACTED>> so can we use the computed property
        attributes = {
            k: v
            for k, v in status.get("attributes", {}).items()
            if v != sensitive_str.redacted_str
        }
        instance = ctor(name, attributes, parent, template, operational)
        if "created" in status:
            instance.created = status["created"]
        if imported:
            self.imports.setShadow(importName, instance)
        return instance

    def findRepoFromGitUrl(self, path, isFile=None, importLoader=None, willUse=False):
        # XXX this doesn't use the tosca template's repositories at all
        repoURL = None
        if not isFile or isURLorGitPath(path):
            # only support urls to git repos for now
            repoURL, filePath, revision = splitGitUrl(path)
        if not repoURL or not self.localEnv:
            return None, None, None, None
        # explicitRevision = revision
        basePath = getBaseDir(importLoader.path)  # checks if dir or not
        # if not explicitRevision:
        #   revision = self.repoStatus.get(repoURL)
        repo, revision, bare = self.localEnv.findOrCreateWorkingDir(
            repoURL, revision, basePath
        )

        # if willUse and (not explicitRevision or not bare):
        #   self.updateRepoStatus(repo, revision) # record that this revision is in use
        return repo, filePath, revision, bare

    def findPathInRepos(self, path, importLoader=None, willUse=False):
        """
    Check if the file path is inside a folder that is managed by a repository.
    If the revision is pinned and doesn't match the repo, it might be bare
    """
        if self.localEnv:
            return self.localEnv.findPathInRepos(path, importLoader)
        elif self.repo:
            repo = self.repo
            filePath = repo.findRepoPath(path)
            if filePath is not None:
                return repo, filePath, repo.revision, False
        # XXX
        # if willUse:
        #     self.updateRepoStatus(repo, revision)
        return None, None, None, None

    # def updateRepoStatus(self, repos):
    #   self.repoStatus.update({repo['url'] : repo['commit']
    #       for repo in repos.values() if repo.get('commit')})

    @staticmethod
    def _getRepositories(tpl):
        return ((tpl.get("spec") or {}).get("service_template") or {}).get(
            "repositories"
        ) or {}

    def _setRepositories(self, tpl):
        # handle missing and none
        st = tpl.get("service_template")
        if not st:
            st = CommentedMap()
            tpl["service_template"] = st
        repositories = st.get("repositories")
        if not repositories:
            repositories = CommentedMap()
            st["repositories"] = repositories

        if "unfurl" not in repositories:
            # add a repository that points to this package
            repositories["unfurl"] = dict(url="file:" + _basepath)
        if "self" not in repositories:
            # this is called too early to use self.getBaseDir()
            path = getBaseDir(self.path) if self.path else "."
            if self.repo:
                url = self.repo.getGitLocalUrl(path, "self")
            else:
                url = "file:" + path
            repositories["self"] = dict(url=url)
        if "spec" not in repositories:
            # if not found assume it points the project root or self if not in a project
            if (
                self.localEnv
                and self.localEnv.project
                and self.localEnv.project is not self.localEnv.homeProject
            ):
                if self.localEnv.project.projectRepo:
                    assert self.localEnv.project.projectRoot
                    url = self.localEnv.project.projectRepo.getGitLocalUrl(
                        self.localEnv.project.projectRoot, "spec"
                    )
                else:
                    url = "file:" + self.localEnv.project.projectRoot
                repositories["spec"] = dict(url=url)
            else:
                repositories["spec"] = repositories["self"]
        return repositories

    def loadYamlInclude(
        self, yamlConfig, templatePath, baseDir, warnWhenNotFound=False
    ):
        """
        This is called while the YAML config is being loaded.
        Returns (url or fullpath, parsed yaml)
        """

        # make a copy because we don't want to update the configuration while
        # in the middle of loading and merging
        repositories = self._getRepositories(yamlConfig.config)
        tmp = {"service_template": {"repositories": repositories.copy()}}
        repositories = self._setRepositories(tmp)

        if isinstance(templatePath, dict):
            artifactTpl = templatePath.copy()
            path = artifactTpl["file"]
            repo = artifactTpl.get("repository")
            if isinstance(repo, dict):
                # a full repository spec maybe part of the include
                reponame = repo.pop("name", os.path.basename(path))
                # replace spec with just its name
                artifactTpl["repository"] = reponame
                repositories[reponame] = repo
            artifact = Artifact(artifactTpl, path=baseDir)
        else:
            artifact = Artifact(dict(file=templatePath), path=baseDir)

        # loadYamlFromArtifact constructs a dummy TOSCA Import loader so we can invoke its URL resolution mechanism
        context = CommentedMap()
        context["repositories"] = repositories
        context.manifest = self
        context.ignoreFileNotFound = warnWhenNotFound
        return loadYamlFromArtifact(context, artifact, yaml)

    def statusSummary(self):
        def summary(instance, indent):
            print(" " * indent, instance)
            indent += 4
            for child in instance.instances:
                summary(child, indent)

        summary(self.rootResource, 0)


class SnapShotManifest(Manifest):
    def __init__(self, manifest, commitId):
        super(SnapShotManifest, self).__init__(manifest.path, self.localEnv)
        self.commitId = commitId
        oldManifest = manifest.repo.show(manifest.path, commitId)
        self.repo = manifest.repo
        self.localEnv = manifest.localEnv
        self.manifest = YamlConfig(
            oldManifest, manifest.path, loadHook=self.loadYamlInclude
        )
        expanded = self.manifest.expanded
        # just needs the spec, not root resource
        self._setSpec(expanded.get("spec", {}))
        self._ready(None)



import unittest
from unfurl.runtime import Status, Priority, NodeInstance, OperationalInstance
from unfurl.job import JobOptions, Runner
from unfurl.configurator import Configurator, ConfigurationSpec
from unfurl.yamlmanifest import YamlManifest
from unfurl.yamlloader import YamlConfig
from unfurl.util import (
    UnfurlError,
    UnfurlValidationError,
    lookupClass,
    API_VERSION,
    sensitive_str,
)
from unfurl.merge import (
    expandDoc,
    restoreIncludes,
    diffDicts,
    mergeDicts,
    patchDict,
    parseMergeKey,
)
from ruamel.yaml.comments import CommentedMap
import traceback
import six
import datetime
import copy
import os.path
import pickle
from click.testing import CliRunner
import logging
from unfurl import initLogging


class SimpleConfigurator(Configurator):
    def run(self, task):
        assert self.canRun(task)
        yield task.done(True, Status.ok)


simpleConfigSpec = ConfigurationSpec("subtask", "instantiate", "SimpleConfigurator", 0)


class TestSubtaskConfigurator(Configurator):
    def run(self, task):
        assert self.canRun(task)
        configuration = yield task.createSubTask(simpleConfigSpec)
        assert configuration.status == Status.ok, configuration.status
        # print ("running TestSubtaskConfigurator")
        yield task.done(True, Status.ok)


class ExpandDocTest(unittest.TestCase):
    doc = {
        "t1": {"b": 2},
        "t2": {"a": {"b": 1}, "c": "c"},
        "t3": "val",
        "t4": ["a", "b"],
        "test1": CommentedMap(
            [("+/t2", None), ("a", {"+/t1": None}), ("d", {"+/t3": None}), ("e", "e")]
        ),
        "test2": [1, {"+/t4": ""}, "+t4", {"+/t4": None}],
    }

    expected = {
        "test1": {"a": {"b": 2}, "c": "c", "d": "val", "e": "e"},
        "test2": [1, "a", "b", "+t4", "a", "b"],
    }

    def test_expandDoc(self):
        includes, expanded = expandDoc(self.doc, cls=CommentedMap)
        self.assertEqual(
            includes,
            {
                ("test1",): [(parseMergeKey("+/t2"), None)],
                ("test1", "a"): [(parseMergeKey("+/t1"), None)],
                ("test1", "d"): [(parseMergeKey("+/t3"), None)],
                ("test2", 1): [(parseMergeKey("+/t4"), "")],
                ("test2", 3): [(parseMergeKey("+/t4"), None)],
            },
        )
        self.assertEqual(expanded["test1"], self.expected["test1"])
        self.assertEqual(expanded["test2"], self.expected["test2"])
        restoreIncludes(includes, self.doc, expanded, CommentedMap)
        # restoreInclude should make expanded look like self.doc
        self.assertEqual(expanded["test1"], self.doc["test1"])
        # XXX restoring lists not implemented:
        # self.assertEqual(expanded["test2"], self.doc["test2"])

    def test_diff(self):
        expectedOld = {"a": 1, "b": {"b1": 1, "b2": 1}, "d": 1}
        old = copy.copy(expectedOld)
        new = {"a": 1, "b": {"b1": 2, "b2": 1}, "c": 2}
        diff = diffDicts(old, new)
        self.assertEqual(diff, {"b": {"b1": 2}, "c": 2, "d": {"+%": "delete"}})
        newNew = mergeDicts(old, diff)
        self.assertEqual(newNew, new)
        self.assertEqual(old, expectedOld)
        patchDict(old, new)
        self.assertEqual(old, new)

    def test_missingInclude(self):
        doc1 = CommentedMap(
            [("+/a/c", None), ("a", {"+/b": None}), ("b", {"c": {"d": 1}})]
        )
        includes, expanded = expandDoc(doc1, cls=CommentedMap)
        self.assertEqual(expanded, {"d": 1, "a": {"c": {"d": 1}}, "b": {"c": {"d": 1}}})
        self.assertEqual(len(includes), 2)

        doc1["missing"] = {"+/include-missing": None}
        with self.assertRaises(UnfurlError) as err:
            includes, expanded = expandDoc(doc1, cls=CommentedMap)

    def test_recursion(self):
        doc = {"test3": {"a": {"recurse": {"+/test3": None}}}}
        with self.assertRaises(UnfurlError) as err:
            includes, expanded = expandDoc(doc)
        self.assertEqual(
            str(err.exception),
            """recursive include "('test3',)" in "('test3', 'a', 'recurse')" when including +/test3""",
        )

        doc2 = {"test4": {"+../test4": None, "child": {}}}
        with self.assertRaises(UnfurlError) as err:
            includes, expanded = expandDoc(doc2)
        self.assertEqual(
            str(err.exception),
            """recursive include "['test4']" in "('test4',)" when including +../test4""",
        )


class JobTest(unittest.TestCase):
    def test_lookupClass(self):
        assert lookupClass("SimpleConfigurator") is SimpleConfigurator

    # XXX rewrite
    # def test_runner(self):
    #   rootResource = NodeInstance('root')
    #   configurationSpec = ConfigurationSpec('test', 'instantiate', 'TestSubtaskConfigurator', 0)
    #   specs = [configurationSpec]
    #   runner = Runner(Manifest(rootResource, specs))
    #   job = runner.run(JobOptions(add=True))
    #   assert not job.unexpectedAbort, job.unexpectedAbort.stackInfo


class PickleTest(unittest.TestCase):
    @unittest.skipIf(six.PY2, "XXX fix pickling in 2.7")
    def test_pickle(self):
        path = __file__ + "/../examples/helm-manifest.yaml"
        manifest = YamlManifest(path=path)
        pickled = pickle.dumps(manifest, -1)
        manifest2 = pickle.loads(pickled)
        # self.assertEqual(manifest.rootResource, manifest2.rootResource)


class OperationalInstanceTest(unittest.TestCase):
    def test_aggregate(self):
        ignoredError = OperationalInstance("error", "ignore")
        assert ignoredError.status == Status.error
        assert ignoredError.priority == Priority.ignore
        assert not ignoredError.operational
        assert not ignoredError.required

        requiredError = OperationalInstance("error", "required")
        assert requiredError.status == Status.error
        assert requiredError.priority == Priority.required
        assert requiredError.required
        assert not requiredError.operational

        aggregateOK = OperationalInstance(Status.ok)
        aggregateOK.dependencies = [ignoredError]
        assert aggregateOK.status == Status.ok

        aggregateError = OperationalInstance(Status.ok)
        aggregateError.dependencies = [ignoredError, requiredError]
        self.assertEqual(aggregateError.status, Status.error)

        aggregateError = OperationalInstance(Status.unknown)
        aggregateError.dependencies = [ignoredError, requiredError]
        self.assertEqual(aggregateError.status, Status.unknown)

        aggregateError = OperationalInstance(Status.ok)
        aggregateError.dependencies = [OperationalInstance("unknown", "optional")]
        self.assertEqual(aggregateError.status, Status.degraded)

        aggregateError = OperationalInstance(Status.unknown)
        aggregateError.dependencies = []
        self.assertEqual(aggregateError.status, Status.unknown)

        aggregateError = OperationalInstance(Status.ok)
        aggregateError.dependencies = []
        self.assertEqual(aggregateError.status, Status.ok)


class RunTest(unittest.TestCase):
    """
  XXX:
  test status: report last state, report config changes, report plan
  test discover
  test failing configurations with abort, continue, revert (both capable of reverting and not)
  test incomplete runs
  test unexpected errors, terminations
  test locking
  test required metadata on resources
  """

    def test_manifest(self):
        simple = (
            """
    apiVersion: %s
    kind: Manifest
    spec:
      instances:
        anInstance:
          # template: foo
          interfaces:
            Standard:
              operations:
               configure:
                implementation:    TestSubtaskConfigurator
                inputs: {}
    """
            % API_VERSION
        )
        manifest = YamlManifest(simple)
        runner = Runner(manifest)
        self.assertEqual(runner.taskCount, 0)
        output = six.StringIO()
        job = runner.run(JobOptions(add=True, out=output, startTime="test"))
        assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
        # workDone includes subtasks
        assert len(job.workDone) == 2, job.workDone

        # manifest shouldn't have changed
        manifest2 = YamlManifest(output.getvalue())
        self.assertEqual(
            manifest2.lastJob["summary"],
            "2 tasks (2 changed, 2 ok, 0 failed, 0 unknown, 0 skipped)",
        )
        output2 = six.StringIO()
        job2 = Runner(manifest2).run(JobOptions(add=True, out=output2))
        # print('2', output2.getvalue())
        assert not job2.unexpectedAbort, job2.unexpectedAbort.getStackTrace()

        # should not find any tasks to run
        assert len(job2.workDone) == 0, job2.workDone
        self.maxDiff = None
        self.assertEqual(output.getvalue(), output2.getvalue())

    def test_template_inheritance(self):
        manifest = """
apiVersion: unfurl/v1alpha1
kind: Manifest
configurators:
  step1:
    spec:
      className: "TestSubtaskConfigurator"
      majorVersion: 0
templates:
  base:
    configurations:
      step1:
        +/configurators/step1:
root:
  instances:
    cloud3: #key is resource name
      +/templates/base:
      +/templates/production:
"""
        with self.assertRaises(UnfurlError) as err:
            YamlManifest(manifest)
        self.assertIn("missing includes: [+/templates/production]", str(err.exception))

    def test_sensitive_logging(self):
        try:
            _logHandler, oldLevel = initLogging(logging.DEBUG)
            oldStream = _logHandler.stream
            _logHandler.stream = six.StringIO()

            logger = logging.getLogger("unfurl")
            logger.info("test1 %s", sensitive_str("sensitive"))

            logger = logging.getLogger("another")
            logger.debug("test2 %s", sensitive_str("sensitive"))

            logOutput = _logHandler.stream.getvalue()
            self.assertIn(sensitive_str.redacted_str, logOutput)
            self.assertNotIn("sensitive", logOutput)
        finally:
            _logHandler.stream = oldStream
            initLogging(oldLevel)


class TestInterface:
    def __init__(self, interfaceName, resource):
        self.name = interfaceName
        self.resource = resource


class InterfaceTest(unittest.TestCase):
    """
  .interfaces:
    inherit: ClassName
    default:
    filter:
  """

    def test_interface(self):
        r = NodeInstance("test")
        r.addInterface(TestInterface)
        className = __name__ + ".TestInterface"
        self.assertEqual(r.attributes[".interfaces"], {className: className})
        i = r.getInterface(className)
        assert i, "interface not found"
        self.assertIs(r, i.resource)
        self.assertEqual(i.name, className)
        self.assertIs(r._interfaces[className], i)

    # XXX test ::container[.interfaces=Container]


class FileTestConfigurator(Configurator):
    def run(self, task):
        assert self.canRun(task)
        assert task.target.attributes["file"] == "foo.txt", task.target.attributes
        assert os.path.isabs(task.inputs["path"]), task.inputs
        assert task.inputs["contents"] == "test", task.inputs["contents"]

        filevalue = task.query({"ref": {"file": "foo.txt"}}, wantList=True)
        assert filevalue._attributes[0].external.type == "file"

        filevalue = task.query(
            {"ref": {"file": "foo.txt"}, "foreach": "path"}, wantList=True
        )
        assert filevalue._attributes[0].external.type == "file"

        value = task.query({"ref": {"file": "foo.txt"}})
        yield task.done(True, Status.ok, result=value)


class FileTest(unittest.TestCase):
    def test_fileRef(self):
        simple = (
            """
    apiVersion: %s
    kind: Manifest
    spec:
      service_template:
        topology_template:
          node_templates:
            test:
              type: tosca.nodes.Root
              properties:
                file:
                  eval:
                    file: foo.txt
              interfaces:
                Standard:
                    create:
                      implementation:  FileTestConfigurator
                      inputs:
                        path:
                          ref: file::path
                        contents:
                          ref: file::contents
    """
            % API_VERSION
        )
        manifest = YamlManifest(simple)
        runner = Runner(manifest)
        output = six.StringIO()
        cliRunner = CliRunner()
        with cliRunner.isolated_filesystem():  # as tmpDir
            with open("foo.txt", "w") as f:
                f.write("test")
            job = runner.run(JobOptions(add=True, out=output, startTime="test"))
        task = list(job.workDone.values())[0]
        self.assertEqual(task.result.result, "foo.txt")

    def test_template_includes(self):
        template = """
apiVersion: unfurl/v1alpha1
kind: Manifest
dsl:
  bar: &bar
    c: 4
spec:
  a: 1
  b: 2
status: {}
    """
        cliRunner = CliRunner()
        with cliRunner.isolated_filesystem():  # as tmpDir
            with open("template.yaml", "w") as f:
                f.write(template)

            instanceYaml = """
apiVersion: unfurl/v1alpha1
kind: Manifest
dsl:
  foo: &foo
    d: 5
  +./bar:
  +./foo:
+include:
  file: template.yaml
+include2:
  file: template.yaml
+?include: missing.yaml
+?include2: missing.yaml
spec:
  +*foo:
  +*bar:
  b: 3
      """

            manifest = YamlManifest(instanceYaml)
            assert manifest.manifest.expanded["dsl"]["c"] == 4
            assert manifest.manifest.expanded["dsl"]["d"] == 5
            assert manifest.manifest.expanded["spec"]["a"] == 1
            assert manifest.manifest.expanded["spec"]["b"] == 3
            assert manifest.manifest.expanded["spec"]["c"] == 4
            assert manifest.manifest.expanded["spec"]["d"] == 5
            # XXX these shouldn't be in expanded:
            # assert "+include" not in manifest.manifest.expanded
            # assert "+?include" not in manifest.manifest.expanded
            assert manifest.manifest.config["dsl"]["foo"].anchor.value == "foo"
            assert manifest.manifest.config["dsl"]["foo"].anchor.always_dump

            output = six.StringIO()
            manifest.dump(output)
            config = YamlConfig(output.getvalue())
            assert config.config["+include"] == {"file": "template.yaml"}
            assert config.config["+include2"] == {"file": "template.yaml"}
            assert config.config["+?include"] == "missing.yaml"
            assert config.config["+?include2"] == "missing.yaml"
            assert "a" not in config.config["spec"]

    # XXX
    # def test_change(self):
    #     """
    # config parameter: file.path
    # run...
    # touch file...
    # run again...
    # assert it triggers update
    # """
    #


class ImportTestConfigurator(Configurator):
    def run(self, task):
        assert self.canRun(task)
        assert task.target.attributes["test"]
        assert task.target.attributes["mapped1"]
        yield task.done(True, Status.ok)


class ImportTest(unittest.TestCase):
    def test_import(self):
        foreign = (
            """
    apiVersion: %s
    kind: Manifest
    spec:
     instances:
      foreign:
        attributes:
          prop1: ok
          prop2: not-a-number
    """
            % API_VERSION
        )
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("foreignmanifest.yaml", "w") as f:
                f.write(foreign)

            importer = (
                """
apiVersion: %s
kind: Manifest
context:
 external:
  test:
    manifest:
      file: foreignmanifest.yaml
    instance: foreign # default is root
    # attributes: # queries into resource
    properties: # expected schema for attributes
      prop2:
       type: number
      prop3:
        type: string
        default: 'default'
spec:
  instances:
    importer:
  service_template:
    topology_template:
      node_templates:
        importer:
          type: tosca.nodes.Root
          properties:
            test:
              eval:
                external: test
            mapped1:
              eval:
                external: test
              foreach: prop1
            mapped2:
              eval:
                external: test
              foreach: prop2
            mapped3:
              eval:
                external: test
              foreach: prop3
          interfaces:
            Standard:
                create: ImportTestConfigurator
      """
                % API_VERSION
            )
            manifest = YamlManifest(importer)
            root = manifest.getRootResource()
            importerResource = root.findResource("importer")
            # assert importing.attributes['test']
            assert root.imports["test"]
            self.assertEqual(importerResource.attributes["mapped1"], "ok")
            with self.assertRaises(UnfurlValidationError) as err:
                importerResource.attributes["mapped2"]
            self.assertIn("schema validation failed", str(err.exception))
            self.assertEqual(importerResource.attributes["mapped3"], "default")
            job = Runner(manifest).run(JobOptions(add=True, startTime="time-to-test"))
            assert job.status == Status.ok, job.summary()



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions
from unfurl.configurator import Configurator, Status
from unfurl.merge import lookupPath
import datetime

manifest = """
apiVersion: unfurl/v1alpha1
kind: Manifest
configurations:
  create:
    implementation:
      className: unfurl.configurators.shell.ShellConfigurator
      environment:
        FOO: "{{inputs.foo}}"
    inputs:
       # test that self-references works in jinja2 templates
       command: "echo ${{inputs.envvar}}"
       timeout: 9999
       foo:     helloworld
       envvar:  FOO
       resultTemplate: |
         - name: SELF
           attributes:
             stdout: "{{ stdout | trim }}"
spec:
  service_template:
    topology_template:
      node_templates:
        test1:
          type: tosca.nodes.Root
          interfaces:
            Standard:
              +/configurations:
"""


class ShellConfiguratorTest(unittest.TestCase):
    def test_shell(self):
        """
    test that runner figures out the proper tasks to run
    """
        runner = Runner(YamlManifest(manifest))

        run1 = runner.run(JobOptions(instance="test1"))
        assert len(run1.workDone) == 1, run1.workDone
        self.assertEqual(
            runner.manifest.getRootResource()
            .findResource("test1")
            .attributes["stdout"],
            "helloworld",
        )
        assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()



"""
lock:
  # XXX Pipfile.lock
  # .tool-versions
  runtime:
    digest:
  spec:
    digest:
  repositories:
     - name:
       digest:
  artifacts:
    - name:
      instance:
      digest:
      changeId:
  ensembles:
    name: # e.g. localhost
      uri:
      changeId:
"""


def saveLock(task, lock):
    instance = task.target
    artifact = instance.artifacts.get("image")
    if artifact:
        name = task.outputs.get("image_path")
        docker_container = task.outputs.get("docker_container")
        if docker_container:
            digest = docker_container["Image"]
            lock.artifacts[instance.name]["image"] = dict(
                name=name, digest=digest, changeId=task.changeId
            )



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions, Status
from unfurl.configurator import Configurator
from unfurl.merge import lookupPath
import datetime


class TestConfigurator(Configurator):
    def run(self, task):
        assert self.canRun(task)
        attributes = task.target.attributes
        attributes["copyOfMeetsTheRequirement"] = attributes["meetsTheRequirement"]
        params = attributes
        if params.get("addresources"):
            shouldYield = params.get("yieldresources")
            resourceSpec = {"name": params["resourceName"], "template": "test1"}
            if shouldYield:
                resourceSpec["dependent"] = True

            updateSpec = dict(name=".self", status=dict(attributes={"newAttribute": 1}))
            task.logger.info("updateResources: %s", [resourceSpec, updateSpec])
            jobrequest = task.updateResources([resourceSpec, updateSpec])
            if shouldYield:
                job = yield jobrequest
                assert job, jobrequest

        yield task.done(True)


manifest = """
apiVersion: unfurl/v1alpha1
kind: Manifest
spec:
 instances:
    test1:
        attributes:
          meetsTheRequirement: "copy"
        interfaces:
          Standard:
            operations:
              configure:
                implementation:
                  className: TestConfigurator
                  majorVersion: 0
                  preConditions:
                    properties:
                      meetsTheRequirement:
                        type: string
                    required: ['meetsTheRequirement']
    test2:
        attributes:
          meetsTheRequirement: false
        +/spec/instances/test1:
    test3:
        +/spec/instances/test1:
        properties:
          resourceName: added1
          addresources: true
    test4:
        +/spec/instances/test3:
        properties:
          resourceName: added2
          yieldresources: true
"""


class ConfiguratorTest(unittest.TestCase):
    def verifyRoundtrip(self, original, jobOptions):
        if jobOptions.startTime:
            jobOptions.startTime += 1
        job = Runner(YamlManifest(original)).run(jobOptions)
        # should not need to run any tasks
        assert len(job.workDone) == 0, job.workDone

        self.maxDiff = None
        self.assertEqual(original, job.out.getvalue())

    def test_neededTasks(self):
        """
    test that runner figures out the proper tasks to run
    """
        runner = Runner(YamlManifest(manifest))
        test1 = runner.manifest.getRootResource().findResource("test1")
        assert test1
        # missing = runner.manifest.spec[0].findInvalidPreconditions(test1)
        # assert not missing, missing

        run1 = runner.run(JobOptions(instance="test1"))
        assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()
        assert len(run1.workDone) == 1, run1.summary()

    def test_preConditions(self):
        """test that the configuration only runs if the resource meets the requirements"""
        runner = Runner(YamlManifest(manifest))
        test1 = runner.manifest.getRootResource().findResource("test1")
        assert test1

        self.assertEqual(test1.attributes["meetsTheRequirement"], "copy")
        #     notYetProvided = runner.manifest.spec[0].findMissingProvided(test1)
        #     self.assertEqual(str(notYetProvided[1]),
        # """[<ValidationError: "'copyOfMeetsTheRequirement' is a required property">]""")

        jobOptions1 = JobOptions(resource="test1", startTime=1)
        run1 = runner.run(jobOptions1)
        # print(run1.out.getvalue())
        assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()
        self.assertEqual(test1.attributes["copyOfMeetsTheRequirement"], "copy")

        # provided = runner.manifest.spec[0].findMissingProvided(test1)
        # assert not provided, provided

        # check that the modifications were recorded
        self.assertEqual(
            runner.manifest.manifest.config["changes"][0]["changes"],
            {"::test1": {"copyOfMeetsTheRequirement": "copy"}},
        )

        test2 = runner.manifest.getRootResource().findResource("test2")
        assert test2
        requiredAttribute = test2.attributes["meetsTheRequirement"]
        assert requiredAttribute is False, requiredAttribute
        # missing = runner.manifest.specs[0].findMissingRequirements(test2)
        # self.assertEqual(str(missing[1]), '''[<ValidationError: "False is not of type 'string'">]''')

        self.verifyRoundtrip(run1.out.getvalue(), jobOptions1)

        jobOptions2 = JobOptions(resource="test2", startTime=2)
        run2 = runner.run(jobOptions2)
        assert not run2.unexpectedAbort, run2.unexpectedAbort.getStackTrace()
        assert run2.status == Status.error, run2.status
        # XXX better error reporting
        # self.assertEqual(str(run2.problems), "can't run required configuration: resource test2 doesn't meet requirement")

        # print(run2.out.getvalue())
        # don't re-run the failed configurations so nothing will have changed
        jobOptions2.repair = "none"
        self.verifyRoundtrip(run2.out.getvalue(), jobOptions2)

    def test_addingResources(self):
        runner = Runner(YamlManifest(manifest))
        jobOptions = JobOptions(resource="test3", startTime=1)
        run = runner.run(jobOptions)
        assert not run.unexpectedAbort, run.unexpectedAbort.getStackTrace()
        # self.assertEqual(list(run.workDone.keys()), [('test3', 'test'), ('added1', 'config1')])

        # print('config', run.out.getvalue())

        changes = runner.manifest.manifest.config["changes"][0]["changes"]
        added = {".added": {"name": "added1", "template": "test1"}}
        self.assertEqual(changes["::added1"], added)

        # verify modified
        self.assertEqual(changes["::test3"], {"copyOfMeetsTheRequirement": "copy"})

        # print('test3', run.out.getvalue())
        jobOptions.repair = "none"
        self.verifyRoundtrip(run.out.getvalue(), jobOptions)

        jobOptions = JobOptions(resource="test4", startTime=2)
        run = runner.run(jobOptions)
        assert not run.unexpectedAbort, run.unexpectedAbort.getStackTrace()
        self.assertEqual(
            [(t.name, t.target.name) for t in run.workDone.values()],
            [
                ("for add: Standard.configure", "test4"),
                ("for add: Standard.configure", "added2"),
            ],
        )
        # print('test4', run.out.getvalue())

        # XXX
        # verify dependencies added
        # dependencies = lookupPath(
        #     runner.manifest.manifest.config,
        #     "root.instances.test4.status.configurations.test.dependencies".split("."),
        # )
        # self.assertEqual(dependencies, [{"ref": "::added2"}])

        jobOptions.repair = "none"
        self.verifyRoundtrip(run.out.getvalue(), jobOptions)

    def test_TemplateConfigurator(self):
        manifest = """\
  apiVersion: unfurl/v1alpha1
  kind: Manifest
  spec:
    service_template:
      topology_template:
        node_templates:
          testNode:
            type: tosca.nodes.Root
            interfaces:
             Standard:
              operations:
                configure:
                  implementation:
                    className: unfurl.configurators.TemplateConfigurator
                  inputs:
                    done:
                      result:
                        outputVar: true
                    resultTemplate: |
                      - name: .self
                        attributes:
                          outputVar: {{ outputVar }}
  """
        runner = Runner(YamlManifest(manifest))
        job = runner.run()
        assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
        self.assertEqual(
            job.stats(),
            {"total": 1, "ok": 1, "error": 0, "unknown": 0, "skipped": 0, "changed": 1},
        )
        assert job.rootResource.findResource("testNode").attributes["outputVar"]

    # def test_shouldRun(self):
    #   pass
    #   #assert should_run
    #   # run()
    #   #assert not should_run
    #
    # def test_provides(self):
    #   #test that it provides as expected
    #   #test that there's an error if provides fails
    #   pass
    #
    # def test_update(self):
    #   #test version changed
    #   pass
    #
    # def test_configChanged(self):
    #   #test version changed
    #   pass
    #
    # def test_revert(self):
    #   # assert provides is removed
    #   pass



from ..util import saveToFile, UnfurlTaskError
from .shell import ShellConfigurator, which
from ..support import getdir, abspath, Status, writeFile
import json
import os
import os.path
import six
import re


def _getEnv(env, verbose, dataDir):
    env["TF_IN_AUTOMATION"] = "1"
    env["TF_INPUT"] = "0"
    # terraform currently only supports TF_LOG=TRACE
    # env["TF_LOG"] = "ERROR WARN INFO DEBUG TRACE".split()[verbose + 1]
    if verbose > 0:
        env["TF_LOG"] = "TRACE"

    # note: modules with relative paths get confused .terraform isn't child of the config dir
    # contains modules/modules.json and plugins/plugins.json:
    env["TF_DATA_DIR"] = dataDir
    return env


def markBlock(schema, items, task, sensitiveNames):
    blockTypes = schema.get("block_types", {})
    attributes = schema.get("attributes", {})
    for obj in items:
        for name, value in obj.items():
            attributeSchema = attributes.get(name)
            if attributeSchema:
                if attributeSchema.get("sensitive") or name in sensitiveNames:
                    #   mark sensistive
                    obj[name] = task.sensitive(value)
            else:
                if not value:
                    continue
                blockSchema = blockTypes.get(name)
                if blockSchema:
                    # "single", "map", "list", "set"
                    objectType = blockSchema["nesting_mode"]
                    if objectType == "single":
                        markBlock(blockSchema["block"], [value], task, sensitiveNames)
                    elif objectType == "map":
                        markBlock(
                            blockSchema["block"], value.values(), task, sensitiveNames
                        )
                    else:
                        markBlock(blockSchema["block"], value, task, sensitiveNames)


def markSensitive(schemas, state, task, sensitiveNames=()):
    for name, attrs in state["outputs"].items():
        value = attrs["value"]
        if attrs.get("sensitive") or name in sensitiveNames:
            state["outputs"][name]["value"] = task.sensitive(value)

    for resource in state["resources"]:
        provider = resource["provider"]
        type = resource["type"]
        providerSchema = schemas.get(provider) or schemas.get(
            provider.lstrip("provider.")
        )
        if providerSchema:
            schema = providerSchema["resource_schemas"].get(type)
            if schema:
                markBlock(schema["block"], resource["instances"], task, sensitiveNames)
            else:
                task.logger.warning(
                    "resource type '%s' not found in terraform schema", type
                )
        else:
            task.logger.warning("provider '%s' not found in terraform schema", provider)
    return state


class TerraformConfigurator(ShellConfigurator):
    _defaultCmd = "terraform"

    sensitiveNames = ["access_token", "key_material", "password"]

    def canDryRun(self, task):
        return True

    def _initTerraform(self, task, terraform, cwd, env):
        echo = task.verbose > -1
        timeout = task.configSpec.timeout
        cmd = terraform + ["init"]
        result = self.runProcess(cmd, timeout=timeout, env=env, cwd=cwd, echo=echo)
        if self._getStatusFromResult(task, result) != Status.ok:
            return None

        cmd = terraform + "providers schema -json".split(" ")
        result = self.runProcess(cmd, timeout=timeout, env=env, cwd=cwd, echo=False)
        if self._getStatusFromResult(task, result) != Status.ok:
            return None

        try:
            providerSchema = json.loads(result.stdout.strip())["provider_schemas"]
            # XXX add to ensemble "lock" section
            # os.path.join(env['TF_DATA_DIR'], "modules", "modules.json")
            # os.path.join(env['TF_DATA_DIR'], "plugins", "plugins.json")
            return providerSchema
        except:
            task.logger.debug("failed to load provider schema", exc_info=True)
            return None

    def _prepareWorkspace(self, cwd, task):
        """
        In terraform directory:
            Write out tf.json if necessary.
        """
        # generated tf.json get written to as main.unfurl.tmp.tf.json
        # (add *.tmp.* to git ignore)
        main = task.inputs.getCopy("main")
        if main:
            if isinstance(main, six.string_types):
                # assume its HCL
                path = os.path.join(cwd, "main.unfurl.tmp.tf")
            else:
                path = os.path.join(cwd, "main.unfurl.tmp.tf.json")
            ctx = task.inputs.context
            return writeFile(ctx, main, path)
        return None

    def _prepareVars(self, task):
        # XXX .tfvars can be sensitive
        # we need to output the plan and convert it to json to see which variables are marked sensitive
        tfvars = task.inputs.getCopy("tfvars")
        if tfvars:
            if isinstance(tfvars, six.string_types):
                # assume the contents of a tfvars file
                path = "vars.tfvars"
            else:
                path = "vars.tfvars.json"
            ctx = task.inputs.context
            return writeFile(ctx, tfvars, path, "local")
        return None

    def _prepareState(self, task, ctx):
        # the terraform state file is associate with the current instance
        # read the (possible encrypted) version from the repository
        # and write out it as plaintext json into the local directory
        yamlPath = abspath(ctx, "terraform.tfstate.yaml", "home", False)
        jsonPath = abspath(ctx, "terraform.tfstate.json", "local", False)
        if not os.path.exists(jsonPath) and os.path.exists(yamlPath):
            # if exists in home, load and write out state file as json
            with open(yamlPath, "r") as f:
                state = task._manifest.yaml.load(f.read())
            writeFile(ctx, state, jsonPath)
        return jsonPath

    def run(self, task):
        ctx = task.inputs.context
        echo = task.verbose > -1
        # options:
        _, terraform = self._cmd(
            task.inputs.get("command", self._defaultCmd), task.inputs.get("keeplines")
        )
        cwd = os.path.abspath(task.inputs.get("dir") or getdir(ctx, "spec.home"))
        dataDir = os.getenv("TF_DATA_DIR", os.path.join(cwd, ".terraform"))

        # write out any needed files to cwd, eg. main.tf.json
        self._prepareWorkspace(cwd, task)
        # write vars to file in local if necessary
        varfilePath = self._prepareVars(task)
        # write the state file to local if necessary
        statePath = self._prepareState(task, ctx)

        env = _getEnv(task.getEnvironment(False), task.verbose, dataDir)

        ### Load the providers schemas and run terraform init if necessary
        providerSchemaPath = os.path.join(dataDir, "providers-schema.json")
        if os.path.exists(providerSchemaPath):
            with open(providerSchemaPath) as psf:
                providerSchema = json.load(psf)
        else:  # first time
            providerSchema = self._initTerraform(task, terraform, cwd, env)
            if providerSchema:
                saveToFile(providerSchemaPath, providerSchema)
            else:
                raise UnfurlTaskError(task, "terraform init failed")

        # build the command line and run it
        if task.dryRun:
            action = ["plan", "-state=" + statePath, "-detailed-exitcode"]
            if task.configSpec.operation == "delete":
                action.append("-destroy")
        elif task.configSpec.operation == "check":
            action = ["refresh", "-state=" + statePath]
        elif task.configSpec.operation == "delete":
            action = ["destroy", "-auto-approve", "-state=" + statePath]
        elif task.configSpec.workflow == "deploy":
            action = ["apply", "-auto-approve", "-state=" + statePath]
        else:
            raise UnfurlTaskError(
                task, "unexpected operation: " + task.configSpec.operation
            )
        cmd = terraform + action
        if varfilePath:
            cmd.append("-var-file=" + varfilePath)

        result = self.runProcess(
            cmd, timeout=task.configSpec.timeout, env=env, cwd=cwd, echo=echo
        )
        if result.returncode and re.search(r"terraform\s+init", result.stderr):
            # modules or plugins out of date, re-run terraform init
            providerSchema = self._initTerraform(task, terraform, cwd, env)
            if providerSchema:
                saveToFile(providerSchemaPath, providerSchema)
                # try again
                result = self.runProcess(
                    cmd, timeout=task.configSpec.timeout, env=env, cwd=cwd, echo=echo
                )
            else:
                raise UnfurlTaskError(task, "terrform init failed")

        # process the result
        if result.returncode == 2:
            status = Status.ok
        else:
            status = self._getStatusFromResult(task, result)

        if status == Status.ok and not task.dryRun:
            # read state file
            with open(statePath) as sf:
                state = json.load(sf)
            state = markSensitive(providerSchema, state, task, self.sensitiveNames)
            # save state file in home as yaml, encrypting sensitive values
            writeFile(ctx, state, "terraform.tfstate.yaml", "home")
            outputs = {name: attrs["value"] for name, attrs in state["outputs"].items()}
            state.update(result.__dict__)
            self.processResultTemplate(task, state)
        else:
            outputs = None

        # XXX how to tell if modified?
        # requires calling plan with -detailed-exitcode to set modified: 0 no change, 1 error, 2 modified (plan only)
        yield task.done(status == Status.ok, result=result.__dict__, outputs=outputs)


# XXX implement discover:
# terraform import -allow-missing-config {type.name}
# convert resource schemas to TOSCA types?
# see https://www.terraform.io/docs/extend/schemas/schema-types.html#types
# types: string int float64 list set map
# behaviors: Default optional required computed (=> attribute) ForceNew: instance-key, sensitive



from collections import Mapping, MutableSequence, MutableMapping
from datetime import datetime, timedelta

from .merge import diffDicts
from .util import UnfurlError, isSensitive, sensitive, sensitive_dict, sensitive_list


def serializeValue(value, **kw):
    getter = getattr(value, "asRef", None)
    if getter:
        return getter(kw)
    if isinstance(value, Mapping):
        ctor = sensitive_dict if isinstance(value, sensitive) else dict
        return ctor((key, serializeValue(v, **kw)) for key, v in value.items())
    if isinstance(value, (MutableSequence, tuple)):
        ctor = sensitive_list if isinstance(value, sensitive) else list
        return ctor(serializeValue(item, **kw) for item in value)
    else:
        return value


class ResourceRef(object):
    # ABC requires 'parent', and '_resolve'

    def _getProp(self, name):
        if name == ".":
            return self
        elif name == "..":
            return self.parent
        name = name[1:]
        # XXX3 use propmap
        return getattr(self, name)

    def __reflookup__(self, key):
        if not key:
            raise KeyError(key)
        if key[0] == ".":
            return self._getProp(key)

        return self._resolve(key)

    def yieldParents(self):
        "yield self and ancestors starting from self"
        resource = self
        while resource:
            yield resource
            resource = resource.parent

    @property
    def ancestors(self):
        return list(self.yieldParents())

    @property
    def parents(self):
        """list of parents starting from root"""
        return list(reversed(self.ancestors))[:-1]

    @property
    def root(self):
        return self.ancestors[-1]

    @property
    def all(self):
        return self.root._all

    @property
    def templar(self):
        return self.root._templar


class ChangeRecord(object):
    """
    A ChangeRecord represents a job or task in the change log file.
    It consists of a change ID and named attributes.

    A change ID is an identifier with this sequence of 12 characters:
    - "A" serves as a format version identifier
    - 7 alphanumeric characters (0-9, A-Z, and a-z) encoding the date and time the job ran.
    - 4 hexadecimal digits encoding the task id
    """

    EpochStartTime = datetime(2020, 1, 1, tzinfo=None)
    LogAttributes = ("previousId",)
    DateTimeFormat = "%Y-%m-%d-%H-%M-%S-%f"

    def __init__(
        self, jobId=None, startTime=None, taskId=0, previousId=None, parse=None
    ):
        if parse:
            self.parse(parse)
            self.setStartTime(getattr(self, "startTime", startTime))
        else:
            self.setStartTime(startTime)
            self.taskId = taskId
            self.previousId = previousId
            if jobId:
                self.changeId = self.updateChangeId(jobId, taskId)
            else:
                self.changeId = self.makeChangeId(self.startTime, taskId, previousId)

    def setStartTime(self, startTime=None):
        if not startTime:
            self.startTime = datetime.utcnow()
        elif isinstance(startTime, datetime):
            self.startTime = startTime
        elif isinstance(startTime, int):  # helper for deterministic testing
            self.startTime = self.EpochStartTime.replace(hour=startTime)
        else:
            try:
                self.startTime = datetime.strptime(startTime, self.DateTimeFormat)
            except ValueError:
                self.startTime = self.EpochStartTime

    def getStartTime(self):
        return self.startTime.strftime(self.DateTimeFormat)

    def setTaskId(self, taskId):
        self.taskId = taskId
        self.changeId = self.updateChangeId(self.changeId, taskId)

    @staticmethod
    def getJobId(changeId):
        return ChangeRecord.updateChangeId(changeId, 0)

    @staticmethod
    def updateChangeId(changeId, taskId):
        return changeId[:-4] + "{:04x}".format(taskId)

    @staticmethod
    def decode(changeId):
        def _decodeChr(i, c):
            offset = 48 if c < "A" else 55
            val = ord(c) - offset
            return str(val + (2020 if i == 0 else 0))

        return (
            "-".join([_decodeChr(*e) for e in enumerate(changeId[1:7])])
            + "."
            + _decodeChr(7, changeId[7])
        )

    @classmethod
    def makeChangeId(self, timestamp=None, taskid=0, previousId=None):
        b62 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
        if not timestamp:
            timestamp = datetime.utcnow()

        year = timestamp.year - self.EpochStartTime.year  # 2020
        if year < 0:
            raise UnfurlError("changeId timestamp too far in the past: %s" % timestamp)
        if year > len(b62):
            raise UnfurlError(
                "changeId timestamp too far in the future: %s" % timestamp
            )

        # year, month, day, hour, minute, second, wday, yday, dst
        jobIdFragment = "".join([b62[n] for n in timestamp.utctimetuple()[1:6]])
        fraction = b62[timestamp.microsecond // 16200]
        changeId = "A{}{}{}{:04x}".format(b62[year], jobIdFragment, fraction, taskid)

        if previousId:
            if previousId[:8] == changeId[:8]:
                # in case last job started less than 1/62nd of a second ago
                return self.makeChangeId(
                    timestamp + timedelta(milliseconds=16200), taskid, previousId
                )
            if previousId > changeId:
                raise UnfurlError(
                    "New changeId is earlier than the previous changeId: %s (%s) < %s (%s) Is time set correctly?"
                    % (
                        changeId,
                        self.decode(changeId),
                        previousId,
                        self.decode(previousId),
                    )
                )
        return changeId

    def parse(self, log):
        terms = log.split("\t")
        attributes = dict(startTime=None)
        for i, term in enumerate(terms):
            if i == 0:
                self.changeId = term
                self.taskId = int(term[-4:], 16)
            # elif i == 1 and '=' not in term:
            #     self.parentId=ChangeId(term)
            else:
                left, sep, right = term.partition("=")
                attributes[left] = right
        self.__dict__.update(attributes)

    def log(self, attributes=None):
        r"changeid\tkey=value\tkey=value"
        # if self.parentId:
        #     start = [str(self.changeId), str(self.parentId)]
        # else:
        start = [self.changeId]
        terms = start + [
            "{}={}".format(k, getattr(self, k))
            for k in self.LogAttributes
            if getattr(self, k, None) is not None
        ]
        if attributes:
            terms += ["{}={}".format(k, v) for k, v in attributes.items()]
        return "\t".join(terms) + "\n"


class ChangeAware(object):
    def hasChanged(self, changeRecord):
        return False


class ExternalValue(ChangeAware):
    __slots__ = ("type", "key")

    def __init__(self, type, key):
        self.type = type
        self.key = key

    def get(self):
        return self.key

    # XXX def __setstate__

    def __eq__(self, other):
        if isinstance(other, ExternalValue):
            return self.get() == other.get()
        return self.get() == other

    def resolveKey(self, key=None, currentResource=None):
        if key:
            value = self.get()
            getter = getattr(value, "__reflookup__", None)
            if getter:
                return getter(key)
            else:
                return value[key]
        else:
            return self.get()

    def asRef(self, options=None):
        if options and options.get("resolveExternal"):
            return serializeValue(self.get(), **options)
        serialized = {self.type: self.key}
        return {"eval": serialized}


_Deleted = object()


class Result(ChangeAware):
    __slots__ = ("original", "resolved", "external", "select")

    def __init__(self, resolved):
        self.select = ()
        self.original = _Deleted
        if isinstance(resolved, ExternalValue):
            self.resolved = resolved.get()
            assert not isinstance(self.resolved, Result), self.resolved
            self.external = resolved
        else:
            assert not isinstance(resolved, Result), resolved
            self.resolved = resolved
            self.external = None

    def asRef(self, options=None):
        options = options or {}
        if self.external:
            ref = self.external.asRef(options)
            if self.select and not options.get("resolveExternal"):
                ref["foreach"] = "." + "::".join(self.select)
            return ref
        else:
            val = serializeValue(self.resolved, **options)
            return val

    def hasDiff(self):
        if self.original is _Deleted:  # this is a new item
            return True
        else:
            if isinstance(self.resolved, Results):
                return self.resolved.hasDiff()
            else:
                newval = self.asRef()
                if self.original != newval:
                    return True
        return False

    def getDiff(self):
        if isinstance(self.resolved, Results):
            return self.resolved.getDiff()
        else:
            val = self.asRef()
            if not self.external and isinstance(val, Mapping):
                old = serializeValue(self.original)
                if isinstance(old, Mapping):
                    return diffDicts(old, val)
            return val

    def __sensitive__(self):
        if self.external:
            return isSensitive(self.external)
        else:
            return isSensitive(self.resolved)

    def _values(self):
        resolved = self.resolved
        if isinstance(resolved, ResultsList):
            # iterate on list to make sure __getitem__ was called
            return (resolved._attributes[i] for (i, v) in enumerate(resolved))
        elif isinstance(resolved, ResultsMap):
            # use items() to make sure __getitem__ was called
            return (resolved._attributes[k] for (k, v) in resolved.items())
        elif isinstance(resolved, Mapping):
            return (Result(i) for i in resolved.values())
        elif isinstance(resolved, MutableSequence):
            return (Result(i) for i in resolved)
        else:
            return resolved

    def _resolveKey(self, key, currentResource):
        # might return a Result
        if self.external:
            value = self.external.resolveKey(key, currentResource)
        else:
            getter = getattr(self.resolved, "__reflookup__", None)
            if getter:
                value = getter(key)
            else:
                value = self.resolved[key]
        return value

    def project(self, key, ctx):
        # returns a Result
        from .eval import Ref

        value = self._resolveKey(key, ctx._lastResource)
        if isinstance(value, Result):
            result = value
        elif Ref.isRef(value):
            result = Ref(value).resolve(ctx, wantList="result")
            if not result:
                raise KeyError(key)
        else:
            result = Result(value)
        if self.external:
            # if value is an ExternalValue this will overwrite it
            result.external = self.external
            result.select = self.select + (key,)
        return result

    def hasChanged(self, changeset):
        if self.external:
            return self.external.hasChanged(changeset)
        elif isinstance(self.resolved, ChangeAware):
            return self.resolved.hasChanged(changeset)
        else:
            return False

    def __eq__(self, other):
        if isinstance(other, Result):
            return self.resolved == other.resolved
        else:
            return self.resolved == other

    def __repr__(self):
        return "Result(%r, %r, %r)" % (self.resolved, self.external, self.select)


class Results(object):
    """
  Evaluating expressions are not guaranteed to be idempotent (consider quoting)
  and resolving the whole tree up front can lead to evaluations of circular references unless the
  order is carefully chosen. So evaluate lazily and memoize the results.
  This also allows us to track changes to the returned structure.
  """

    __slots__ = ("_attributes", "context", "_deleted")

    doFullResolve = False

    def __init__(self, serializedOriginal, resourceOrCxt):
        from .eval import RefContext

        assert not isinstance(serializedOriginal, Results), serializedOriginal
        self._attributes = serializedOriginal
        self._deleted = {}
        if not isinstance(resourceOrCxt, RefContext):
            ctx = RefContext(resourceOrCxt)
        else:
            ctx = resourceOrCxt

        oldBaseDir = ctx.baseDir
        newBaseDir = getattr(serializedOriginal, "baseDir", oldBaseDir)
        if newBaseDir and newBaseDir != oldBaseDir:
            ctx = ctx.copy()
            ctx.baseDir = newBaseDir
            ctx.trace("found baseDir", newBaseDir, "old", oldBaseDir)
        self.context = ctx

    def getCopy(self, key, default=None):
        from .eval import mapValue

        try:
            val = self._attributes[key]
        except:
            val = default
        else:
            if isinstance(val, Result):
                assert not isinstance(val.resolved, Result), val
                if val.original is _Deleted:
                    val = val.asRef()
                else:
                    val = val.original
        return mapValue(val, self.context)

    @staticmethod
    def _mapValue(val, context):
        "Recursively and lazily resolves any references in a value"
        from .eval import mapValue, Ref

        if isinstance(val, Results):
            return val
        elif Ref.isRef(val):
            return Ref(val).resolve(context, wantList="result")
        elif isinstance(val, sensitive):
            return val
        elif isinstance(val, Mapping):
            return ResultsMap(val, context)
        elif isinstance(val, list):
            return ResultsList(val, context)
        else:
            # at this point, just evaluates templates in strings or returns val
            return mapValue(val, context)

    def __sensitive__(self):
        # only check resolved values
        return any(isinstance(x, Result) and isSensitive(x) for x in self._values())

    def hasDiff(self):
        # only check resolved values
        return any(isinstance(x, Result) and x.hasDiff() for x in self._values())

    def __getitem__(self, key):
        from .eval import mapValue

        val = self._attributes[key]
        if isinstance(val, Result):
            assert not isinstance(val.resolved, Result), val
            return val.resolved
        else:
            if self.doFullResolve:
                if isinstance(val, Results):
                    resolved = val
                else:  # evaluate records that aren't Results
                    resolved = mapValue(val, self.context)
            else:
                # lazily evaluate lists and dicts
                self.context.trace("Results._mapValue", val)
                resolved = self._mapValue(val, self.context)
            # will return a Result if val was an expression that was evaluated
            if isinstance(resolved, Result):
                result = resolved
                result.original = val
                resolved = result.resolved
            else:
                result = Result(resolved)
                result.original = val
            self._attributes[key] = result
            assert not isinstance(resolved, Result), val
            return resolved

    def __setitem__(self, key, value):
        assert not isinstance(value, Result), (key, value)
        self._attributes[key] = Result(value)
        self._deleted.pop(key, None)

    def __delitem__(self, index):
        val = self._attributes[index]
        self._deleted[index] = val
        del self._attributes[index]

    def __len__(self):
        return len(self._attributes)

    def __eq__(self, other):
        if isinstance(other, Results):
            return self._attributes == other._attributes
        else:
            self.resolveAll()
            return self._attributes == other

    def __str__(self):
        return str(self._attributes)

    def __repr__(self):
        return "Results(%r)" % self._attributes


class ResultsMap(Results, MutableMapping):
    def __iter__(self):
        return iter(self._attributes)

    def resolveAll(self):
        list(self.values())

    def serializeResolved(self, **kw):
        return dict(
            (key, serializeValue(v, **kw))
            for key, v in self._attributes.items()
            if isinstance(v, Result)
        )

    def _values(self):
        return self._attributes.values()

    def getDiff(self, cls=dict):
        # returns a dict with the same semantics as diffDicts
        diffDict = cls()
        for key, val in self._attributes.items():
            if isinstance(val, Result) and val.hasDiff():
                diffDict[key] = val.getDiff()

        for key in self._deleted:
            diffDict[key] = {"+%": "delete"}

        return diffDict


class ResultsList(Results, MutableSequence):
    def insert(self, index, value):
        assert not isinstance(value, Result), value
        self._attributes.insert(index, Result(value))

    def _values(self):
        return self._attributes

    def getDiff(self, cls=list):
        # we don't have patchList yet so just returns the whole list
        return cls(
            val.getDiff() if isinstance(val, Result) else val
            for val in self._attributes
        )

    def resolveAll(self):
        list(self)



TOSCA_VERSION = 'tosca_simple_unfurl_1_0_0'

# loaded by toscaparser extention manager using entrypoint declared in setup.cfg
class plugindef_1_0_0(object):
    VERSION = TOSCA_VERSION

    DEFS_FILE = "tosca-ext.yaml"

    SECTIONS = ('metadata')



"""
Creates and clone projects and ensembles.

After running "init" your Unfurl project will look like:

ensemble/ensemble.yaml
ensemble-template.yaml
unfurl.yaml
local/unfurl.yaml

If the --existing option is used, the project will be added to the nearest repository found in a parent folder.
If the --mono option is used, the ensemble add the project repo instead of it's own.

Each repository created will also have .gitignore and .gitattributes added.

When a repository is added as child of another repo, that folder will be added to .git/info/exclude
(instead of .gitignore because they shouldn't be committed into the repository).

Paths are always relative but you can optionally specify which repository a path is relative to.

There are three predefined repositories:

"self", which represents the location the ensemble lives in -- it will be
a "git-local:" URL or a "file:" URL if the ensemble is not part of a git repository.

"unfurl" which points to the Python package of the unfurl process -- this can be used to load configurators and templates
that ship with Unfurl.

"spec" which, unless otherwise specified, points to the project root or the ensemble itself if it is not part of a project.
"""
import uuid
import os
import os.path
import datetime
import sys
import shutil
from . import DefaultNames, getHomeConfigPath
from .tosca import TOSCA_VERSION
from .repo import Repo, GitRepo, splitGitUrl, isURLorGitPath
from .util import UnfurlError, getBaseDir
from .localenv import LocalEnv, Project
import random
import string

_templatePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates")


def renameForBackup(dir):
    ctime = datetime.datetime.fromtimestamp(os.stat(dir).st_ctime)
    new = dir + "." + ctime.strftime("%Y-%m-%d-%H-%M-%S")
    os.rename(dir, new)
    return new


def get_random_password(count=12, prefix="uv"):
    srandom = random.SystemRandom()
    start = string.ascii_letters + string.digits
    source = string.ascii_letters + string.digits + "%&()*+,-./:<>?=@[]^_`{}~"
    return prefix + "".join(
        srandom.choice(source if i else start) for i in range(count)
    )


def _writeFile(folder, filename, content):
    if not os.path.isdir(folder):
        os.makedirs(folder)
    filepath = os.path.join(folder, filename)
    with open(filepath, "w") as f:
        f.write(content)
    return filepath


def writeTemplate(folder, filename, template, vars, templateDir=None):
    from .runtime import NodeInstance
    from .eval import RefContext
    from .support import applyTemplate

    if templateDir and not os.path.isabs(templateDir):
        templateDir = os.path.join(_templatePath, templateDir)
    if not templateDir or not os.path.exists(os.path.join(templateDir, template)):
        templateDir = _templatePath  # default

    with open(os.path.join(templateDir, template)) as f:
        source = f.read()
    instance = NodeInstance()
    instance._baseDir = _templatePath
    content = applyTemplate(source, RefContext(instance, vars))
    return _writeFile(folder, filename, content)


def writeProjectConfig(
    projectdir,
    filename=DefaultNames.LocalConfig,
    templatePath=DefaultNames.LocalConfig + ".j2",
    vars=None,
    templateDir=None,
):
    _vars = dict(include="", manifestPath=None)
    if vars:
        _vars.update(vars)
    return writeTemplate(projectdir, filename, templatePath, _vars, templateDir)


def renderHome(homePath):
    homedir, filename = os.path.split(homePath)
    writeTemplate(homedir, DefaultNames.Ensemble, "manifest.yaml.j2", {}, "home")
    configPath = writeProjectConfig(
        homedir, filename, "unfurl.yaml.j2", templateDir="home"
    )
    writeTemplate(homedir, ".gitignore", "gitignore.j2", {})
    return configPath


def createHome(home=None, render=False, replace=False, **kw):
    """
    Create the home project if missing
    """
    homePath = getHomeConfigPath(home)
    if not homePath:
        return None
    exists = os.path.exists(homePath)
    if exists and not replace:
        return None

    homedir, filename = os.path.split(homePath)
    if render:  # just render
        repo = None
    else:
        if exists:
            renameForBackup(homedir)
        repo = _createRepo(homedir)

    configPath = renderHome(homePath)
    if not render and not kw.get("no_runtime"):
        initEngine(homedir, kw.get("runtime") or "venv:")
    if repo:
        createProjectRepo(homedir, repo, True, addDefaults=False, templateDir="home")
        repo.repo.git.add("--all")
        repo.repo.index.commit("Create the unfurl home repository")
        repo.repo.git.branch("rendered")  # now create a branch
    return configPath


def _createRepo(gitDir, gitUri=None):
    import git

    if not os.path.isdir(gitDir):
        os.makedirs(gitDir)
    repo = git.Repo.init(gitDir)
    repo.index.add(addHiddenGitFiles(gitDir))
    repo.index.commit("Initial Commit for %s" % uuid.uuid1())

    Repo.ignoreDir(gitDir)
    return GitRepo(repo)


def writeServiceTemplate(projectdir):
    vars = dict(version=TOSCA_VERSION)
    return writeTemplate(
        projectdir, "service-template.yaml", "service-template.yaml.j2", vars
    )


def writeEnsembleManifest(
    destDir, manifestName, specRepo, specDir=None, extraVars=None
):
    if extraVars is None:
        # default behaviour is to include the ensembleTemplate
        # in the root of the specDir
        extraVars = dict(ensembleTemplate=DefaultNames.EnsembleTemplate)

    if specDir:
        specDir = os.path.abspath(specDir)
    else:
        specDir = ""
    vars = dict(specRepoUrl=specRepo.getGitLocalUrl(specDir, "spec"))
    vars.update(extraVars)
    return writeTemplate(destDir, manifestName, "manifest.yaml.j2", vars)


def addHiddenGitFiles(gitDir):
    # write .gitignore and  .gitattributes
    gitIgnorePath = writeTemplate(gitDir, ".gitignore", "gitignore.j2", {})
    gitAttributesContent = "**/*%s merge=union\n" % DefaultNames.JobsLog
    gitAttributesPath = _writeFile(gitDir, ".gitattributes", gitAttributesContent)
    return [os.path.abspath(gitIgnorePath), os.path.abspath(gitAttributesPath)]


def createProjectRepo(
    projectdir,
    repo,
    mono,
    addDefaults=True,
    projectConfigTemplate=None,
    templateDir=None,
):
    """
    Creates a folder named `projectdir` with a git repository with the following files:

    unfurl.yaml
    local/unfurl.yaml
    ensemble-template.yaml
    ensemble/ensemble.yaml

    Returns the absolute path to unfurl.yaml
    """
    # write the project files
    localConfigFilename = DefaultNames.LocalConfig
    manifestPath = os.path.join(DefaultNames.EnsembleDirectory, DefaultNames.Ensemble)

    vars = dict(vaultpass=get_random_password())
    # manifestPath should be in local if ensemble is a separate repo
    if addDefaults and not mono:
        vars["manifestPath"] = manifestPath
    writeProjectConfig(
        os.path.join(projectdir, "local"),
        localConfigFilename,
        "unfurl.local.yaml.j2",
        vars,
        templateDir,
    )

    localInclude = "+?include: " + os.path.join("local", localConfigFilename)
    vars = dict(include=localInclude)
    # manifestPath should here if ensemble is part of the repo
    if addDefaults and mono:
        vars["manifestPath"] = manifestPath
    projectConfigPath = writeProjectConfig(
        projectdir,
        DefaultNames.LocalConfig,
        projectConfigTemplate or DefaultNames.LocalConfig + ".j2",
        vars,
        templateDir,
    )
    files = [projectConfigPath]

    if addDefaults:
        # write ensemble-template.yaml
        ensembleTemplatePath = writeTemplate(
            projectdir, DefaultNames.EnsembleTemplate, "manifest-template.yaml.j2", {}
        )
        files.append(ensembleTemplatePath)
        ensembleDir = os.path.join(projectdir, DefaultNames.EnsembleDirectory)
        manifestName = DefaultNames.Ensemble
        if mono:
            ensembleRepo = repo
        else:
            ensembleRepo = _createRepo(ensembleDir)
        extraVars = dict(
            ensembleUri=ensembleRepo.getUrlWithPath(
                os.path.abspath(os.path.join(ensembleDir, manifestName))
            )
        )
        # write ensemble/ensemble.yaml
        manifestPath = writeEnsembleManifest(
            ensembleDir, manifestName, repo, extraVars=extraVars
        )
        if mono:
            files.append(manifestPath)
        else:
            ensembleRepo.repo.index.add([os.path.abspath(manifestPath)])
            ensembleRepo.repo.index.commit("Default ensemble repository boilerplate")

    repo.commitFiles(files, "Create a new unfurl repository")
    return projectConfigPath


def createProject(
    projectdir, home=None, mono=False, existing=False, addDefaults=True, **kw
):
    if existing:
        repo = Repo.findContainingRepo(projectdir)
        if not repo:
            raise UnfurlError("Could not find an existing repository")
    else:
        repo = None
    # creates home if it doesn't exist already:
    newHome = createHome(home, **kw)

    if repo:
        repo.repo.index.add(addHiddenGitFiles(projectdir))
        repo.repo.index.commit("Adding Unfurl project")
    else:
        repo = _createRepo(projectdir)

    # XXX add project to ~/.unfurl_home/unfurl.yaml
    projectConfigPath = createProjectRepo(
        projectdir, repo, mono, addDefaults=addDefaults
    )

    if not newHome and not kw.get("no_runtime") and kw.get("runtime"):
        # if runtime was explicitly set and we aren't creating the home project
        # then initialize the runtime here
        initEngine(projectdir, kw.get("runtime"))

    return newHome, projectConfigPath


def cloneLocalRepos(manifest, sourceProject, targetProject):
    # We need to clone repositories that are local to the source project
    for repoSpec in manifest.repositories.values():
        if repoSpec.name == "self":
            continue
        repo = sourceProject.findRepository(repoSpec)
        if repo:
            targetProject.findOrClone(repo)


def initNewEnsemble(manifest, sourceProject, targetProject):
    cloneLocalRepos(manifest, sourceProject, targetProject)

    destDir = os.path.dirname(manifest.manifest.path)
    if targetProject.projectRepo and targetProject is not sourceProject:
        repo = targetProject.projectRepo
    else:
        # XXX support --mono and --existing
        repo = _createRepo(destDir)

    manifest.metadata["uri"] = repo.getUrlWithPath(manifest.manifest.path)
    with open(manifest.manifest.path, "w") as f:
        manifest.dump(f)

    repo.repo.index.add([manifest.manifest.path])
    repo.repo.index.commit("Default ensemble repository boilerplate")
    return repo


def _looksLike(path, name):
    # in case path is a directory:
    if os.path.isfile(os.path.join(path, name)):
        return path, name
    if path.endswith(name):  # name is explicit so don't need to check if file exists
        return os.path.split(path)
    return None


def _getEnsemblePaths(sourcePath, sourceProject):
    # look for an ensemble-template or service-template in source path
    template = _looksLike(sourcePath, DefaultNames.EnsembleTemplate)
    if template:
        return dict(sourceDir=template[0], ensembleTemplate=template[1])
    template = _looksLike(sourcePath, DefaultNames.ServiceTemplate)
    if template:
        return dict(sourceDir=template[0], serviceTemplate=template[1])
    else:
        # we couldn't find one of the default template files, so treat sourcePath
        # as a path to an ensemble
        # XXX if sourcePath points to an explicit file we should check if it's a service template first
        try:
            localEnv = LocalEnv(sourcePath, project=sourceProject)
            return dict(manifestPath=localEnv.manifestPath, localEnv=localEnv)
        except:
            # nothing found
            return {}


def createNewEnsemble(templateVars, sourceProject, targetPath, targetProject):
    """
    include directives, imports, or external file reference they are guaranteed to be local to the project.
    Outside paths need to be referenced with a named repository.
    """
    # if targetPath is relative, it will be treated as relative to the targetProject root
    from unfurl import yamlmanifest

    if not targetPath:
        destDir, manifestName = DefaultNames.EnsembleDirectory, DefaultNames.Ensemble
        targetPath = os.path.join(destDir, manifestName)
    elif targetPath.endswith(".yaml") or targetPath.endswith(".yml"):
        destDir, manifestName = os.path.split(targetPath)
    else:
        destDir = targetPath
        manifestName = DefaultNames.Ensemble
    # choose a destDir that doesn't conflict with an existing folder
    # (i.e. if default ensemble already exists)
    destDir = targetProject.getUniquePath(destDir)
    targetPath = os.path.join(destDir, manifestName)

    if "manifestPath" not in templateVars:
        # we found a template file to clone
        assert sourceProject
        sourceDir = templateVars["sourceDir"]
        specRepo, relPath, revision, bare = sourceProject.findPathInRepos(sourceDir)
        if not specRepo:
            raise UnfurlError(
                '"%s" is not in a git repository. Cloning from plain file directories not yet supported'
                % sourceDir
            )
        manifestPath = writeEnsembleManifest(
            destDir, manifestName, specRepo, sourceDir, templateVars
        )
        localEnv = LocalEnv(manifestPath, project=sourceProject)
        manifest = yamlmanifest.ReadOnlyManifest(localEnv=localEnv)
    elif templateVars:
        # didn't find a template file
        # look for an ensemble at the given path or use the source project's default
        manifest = yamlmanifest.clone(templateVars["localEnv"], targetPath)
    else:
        return None  # can't find anything to clone
    repo = initNewEnsemble(manifest, sourceProject, targetProject)
    return destDir, repo
    # XXX need to add manifest to unfurl.yaml


def getSourceProject(source, destDir):
    # check if source is a git url
    if isURLorGitPath(source):
        repoURL, filePath, revision = splitGitUrl(source)
        # if destProject add repo to project
        destRoot = Project.findPath(destDir)
        if destRoot:
            # destination is in an existing project, use that one
            sourceProject = Project(destRoot)
            repo = sourceProject.findGitRepo(repoURL, revision)
            if not repo:
                repo = sourceProject.createWorkingDir(repoURL, revision)
            source = os.path.join(repo.workingDir, filePath)
        else:
            # otherwise clone to dest
            Repo.createWorkingDir(repoURL, destDir, revision)
            targetDir = os.path.join(destDir, filePath)
            sourceRoot = Project.findPath(targetDir)
            if sourceRoot:
                sourceProject = Project(sourceRoot)
                source = targetDir
                # we cloned the repo to destDir but the might not be at it's root
                destDir = sourceProject.projectRoot
            else:
                return (
                    None,
                    None,
                    None,
                    (
                        'Error: cloned "%s" to "%s" but couldn\'t find an Unfurl project'
                        % (source, destDir)
                    ),
                )
    else:
        sourceRoot = Project.findPath(source)
        if sourceRoot:
            sourceProject = Project(sourceRoot)
        else:
            # XXX support cloning from an ensemble or template that isn't in a project
            return (
                None,
                None,
                None,
                (
                    "Can't clone \"%s\": it isn't in an Unfurl project or repository"
                    % source
                ),
            )
    return sourceProject, source, destDir, None


def cloneEnsemble(source, destDir, includeLocal=False, **options):
    """
    Clone the `source` ensemble to `dest`. If `dest` isn't in a project, create one.
    `source` can point to an ensemble_template, a service_template, an existing ensemble
    or a folder containing one of those. If it points to a project its default ensemble will be cloned.

    Referenced `repositories` will be cloned if a git repository or copied if a regular file folder,
    If the folders already exist they will be copied to new folder unless the git repositories have the same HEAD.
    but the local repository names will remain the same.
    """
    from unfurl import yamlmanifest

    sourceProject, source, destDir, error = getSourceProject(source, destDir)
    if error:
        return error

    newProject = False
    paths = _getEnsemblePaths(source, sourceProject)
    relDestDir = sourceProject.getRelativePath(destDir)
    # if dest is in the source project:
    if not relDestDir.startswith(".."):
        # add a new ensemble to the current project
        destProject = sourceProject
        destDir = "" if relDestDir == "." else relDestDir
    else:
        destRoot = Project.findPath(destDir)
        if destRoot:
            # destination is in an existing project, use that one
            destProject = Project(destRoot)
        else:  # create new project
            newProject = True
            destProject = None
            # if source in the sourceProject's projectRepo, clone that repo
            # instead of creating a new project repo
            if sourceProject.projectRepo:
                # check if source points to an ensemble that is part of the project repo
                # if source is root, we need to get the default ensemble
                pathToEnsemble = paths.get("manifestPath")
                if pathToEnsemble:
                    srcInRepo = (
                        sourceProject.projectRepo.findRepoPath(pathToEnsemble)
                        is not None
                    )
                    if srcInRepo:
                        # if source is an ensemble and it's in the project repo
                        # we've cloned the ensemble already so we're done
                        sourceProject.projectRepo.clone(destDir)
                        destProject = Project(Project.findPath(destDir))
                        manifest = yamlmanifest.ReadOnlyManifest(
                            localEnv=paths["localEnv"]
                        )
                        cloneLocalRepos(manifest, sourceProject, destProject)
                        return "Cloned project to " + os.path.abspath(destDir)
                    # XXX
                    # else:
                    #     # the source ensemble is not in the project repo
                    #     # if one of the repositories it references is the project repo, clone it
                    #     specFolder = getSpecFoldersFromEnsemble(pathToEnsemble)
                    #     specInRepo = (
                    #         sourceProject.projectRepo.findRepoPath(specFolder)
                    #         is not None
                    #     )
                    #     if specInRepo:
                    #         sourceProject.projectRepo.clone(destDir)
                    #         destProject = Project(destDir)

                # if source is at root of the project repo, just clone it
                if os.path.normpath(sourceProject.projectRepo.workingDir) == getBaseDir(
                    source
                ):
                    # clone the project but we still need to add a new ensemble to it
                    destDir = os.path.abspath(destDir)
                    sourceProject.projectRepo.clone(destDir)
                    # print(
                    #    "cloning", sourceProject.projectRepo.workingDir, "to", destDir
                    # )
                    destProject = Project(Project.findPath(destDir))
                    destDir = DefaultNames.EnsembleDirectory

            if not destProject:
                # create a new project from scratch for the new ensemble
                newHome, projectConfigPath = createProject(
                    destDir, addDefaults=False, **options
                )
                destProject = Project(projectConfigPath)
                destDir = DefaultNames.EnsembleDirectory

    # this will clone the default ensemble if it doesn't point to a specific spec or ensemble
    destDir, repo = createNewEnsemble(paths, sourceProject, destDir, destProject)
    return 'Created new ensemble "%s" in %s project at "%s"' % (
        destDir,
        "new" if newProject else "existing",
        os.path.abspath(destProject.projectRoot),
    )


def initEngine(projectDir, runtime):
    kind, sep, rest = runtime.partition(":")
    if kind == "venv":
        return createVenv(projectDir, rest)
    # elif kind == 'docker'
    # XXX return 'unrecoginized runtime string: "%s"'
    return False


def _addUnfurlToVenv(projectdir):
    # this is hacky
    # can cause confusion if it exposes more packages than unfurl
    base = os.path.dirname(os.path.dirname(_templatePath))
    sitePackageDir = None
    libDir = os.path.join(projectdir, os.path.join(".venv", "lib"))
    for name in os.listdir(libDir):
        sitePackageDir = os.path.join(libDir, name, "site-packages")
        if os.path.isdir(sitePackageDir):
            break
    else:
        # XXX report error: can't find site-packages folder
        return
    _writeFile(sitePackageDir, "unfurl.pth", base)
    _writeFile(sitePackageDir, "unfurl.egg-link", base)


def createVenv(projectDir, pipfileLocation):
    """Create a virtual python environment for the given project."""
    os.environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
    os.environ["PIPENV_VENV_IN_PROJECT"] = "1"
    if "PIPENV_PYTHON" not in os.environ:
        os.environ["PIPENV_PYTHON"] = sys.executable

    try:
        cwd = os.getcwd()
        os.chdir(projectDir)
        # need to set env vars and change current dir before importing pipenv
        from pipenv.core import do_install, ensure_python
        from pipenv.utils import python_version

        pythonPath = str(ensure_python())
        assert pythonPath, pythonPath
        if not pipfileLocation:
            versionStr = python_version(pythonPath)
            assert versionStr, versionStr
            version = versionStr.rpartition(".")[0]  # 3.8.1 => 3.8
            # version = subprocess.run([pythonPath, "-V"]).stdout.decode()[
            #     7:10
            # ]  # e.g. Python 3.8.1 => 3.8
            pipfileLocation = os.path.join(
                _templatePath, "python" + version
            )  # e.g. templates/python3.8

        if not os.path.isdir(pipfileLocation):
            # XXX 'Pipfile location is not a valid directory: "%s" % pipfileLocation'
            return False

        # copy Pipfiles to project root
        if os.path.abspath(projectDir) != os.path.abspath(pipfileLocation):
            for filename in ["Pipfile", "Pipfile.lock"]:
                path = os.path.join(pipfileLocation, filename)
                if os.path.isfile(path):
                    shutil.copy(path, projectDir)

        # create the virtualenv and install the dependencies specified in the Pipefiles
        sys_exit = sys.exit
        try:
            retcode = -1

            def noexit(code):
                retcode = code

            sys.exit = noexit

            do_install(python=pythonPath)
            # this doesn't actually install the unfurl so link to this one
            _addUnfurlToVenv(projectDir)
        finally:
            sys.exit = sys_exit

        return not retcode  # retcode means error
    finally:
        os.chdir(cwd)



import itertools
import re
import six
from collections import Mapping, MutableSequence, namedtuple, Sequence

from ruamel.yaml.comments import CommentedMap, CommentedBase

from .util import UnfurlError


def _mapCtor(self):
    if hasattr(self, "baseDir"):
        return makeMapWithBase(self, self.baseDir)
    return CommentedMap


CommentedMap.mapCtor = property(_mapCtor)


def copy(src):
    cls = getattr(src, "mapCtor", src.__class__)
    if six.PY2 and cls is CommentedMap:
        return CommentedMap(src.items())
    return cls(src)


def makeMapWithBase(doc, baseDir):
    loadTemplate = getattr(doc, "loadTemplate", None)
    _anchorCache = getattr(doc, "_anchorCache", None)

    def factory(*args, **kws):
        if six.PY2 and args and isinstance(args[0], dict):
            map = CommentedMap(args[0].items(), **kws)
        else:
            map = CommentedMap(*args, **kws)
        map.baseDir = baseDir
        if loadTemplate:
            map.loadTemplate = loadTemplate
        if _anchorCache is not None:
            map._anchorCache = _anchorCache
        return map

    return factory


# XXX?? because json keys are strings allow number keys to merge with lists
# other values besides delete not supported because current code can leave those keys in final result
mergeStrategyKey = "+%"  # values: delete

# b is the merge patch, a is original dict
def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
    """
  Returns a new dict (or cls) that recursively merges b into a.
  b is base, a overrides.

  Similar to https://yaml.org/type/merge.html but does a recursive merge
  """
    cls = getattr(b, "mapCtor", cls or b.__class__)
    cp = cls()
    skip = []
    for key, val in a.items():
        if key == mergeStrategyKey:
            continue
        if replaceKeys and key in replaceKeys:
            childStrategy = "replace"
        else:
            childStrategy = "merge"
        if isinstance(val, Mapping):
            strategy = val.get(mergeStrategyKey)
            if key in b:
                bval = b[key]
                if isinstance(bval, Mapping):
                    # merge strategy of a overrides b
                    if not strategy:
                        strategy = bval.get(mergeStrategyKey) or defaultStrategy
                    if strategy == "merge":
                        if not val:  # empty map, treat as missing key
                            continue
                        cp[key] = mergeDicts(
                            bval,
                            val,
                            defaultStrategy=childStrategy,
                            replaceKeys=replaceKeys,
                        )
                        continue
                    if strategy == "error":
                        raise UnfurlError(
                            "merging %s is not allowed, +%: error was set" % key
                        )
                # otherwise we ignore bval because key is already in a
            if strategy == "delete":
                skip.append(key)
                continue
        # XXX merge lists
        # elif isinstance(val, list) and key in b:
        #   bval = b[key]
        #   if isinstance(bval, list):
        #     if appendlists == 'all' or key in appendlists:
        #       cp[key] = bval + val
        #       continue
        #     elif mergelists == 'all' or key in mergelists:
        #       newlist = []
        #       for ai, bi in zip(val, bval):
        #         if isinstance(ai, Mapping) and isinstance(bi, Mapping):
        #           newlist.append(mergeDicts(bi, ai, cls))
        #         elif a1 != deletemarker:
        #           newlist.append(a1)
        #       cp[key] == newlist
        #       continue

        # otherwise a replaces b
        cp[key] = val

    # add new keys
    for key, val in b.items():
        if key == mergeStrategyKey:
            continue
        if key not in cp and key not in skip:
            # note: val is shared not copied
            cp[key] = val
    return cp


def _cacheAnchors(_anchorCache, obj):
    anchor = obj.yaml_anchor()
    if anchor and anchor.value:
        _anchorCache[anchor.value] = obj
        # By default, anchors are only emitted when it detects a reference t
        # to an object previously seen, so force it to be emitted
        anchor.always_dump = True

    for value in obj.values() if isinstance(obj, dict) else obj:
        if isinstance(value, CommentedBase):
            _cacheAnchors(_anchorCache, value)


def findAnchor(doc, anchorName):
    if not isinstance(doc, CommentedMap):
        return None

    _anchorCache = getattr(doc, "_anchorCache", None)
    if _anchorCache is None:
        _anchorCache = {}
        # recursively find anchors
        _cacheAnchors(_anchorCache, doc)
        doc._anchorCache = _anchorCache

    return _anchorCache.get(anchorName)


def _jsonPointerUnescape(s):
    return s.replace("~1", "/").replace("~0", "~")


_RE_INVALID_JSONPOINTER_ESCAPE = re.compile("(~[^01]|~$)")


def _jsonPointerValidate(pointer):
    invalid_escape = _RE_INVALID_JSONPOINTER_ESCAPE.search(pointer)
    if invalid_escape:
        raise UnfurlError(
            "Found invalid escape {} in JSON pointer {}".format(
                invalid_escape.group(), pointer
            )
        )
    return None


def getTemplate(doc, key, value, path, cls):
    template = doc
    templatePath = None
    if key.include:
        value, template, baseDir = doc.loadTemplate(value, key.maybe)
        if template is None:  # include wasn't not found and key.maybe
            return doc
        cls = makeMapWithBase(doc, baseDir)
        # fileKey = key._replace(include=None)
        # template = getTemplate(template, fileKey, "raw", (), cls)
    else:
        result = _findTemplate(doc, key, path, cls, not key.maybe)
        if result is None:
            return doc
        template, templatePath = result
        if templatePath is None:
            templatePath = key.pointer

    try:
        if value != "raw" and isinstance(
            template, Mapping
        ):  # raw means no further processing

            # if the include path starts with the path to the template
            # throw recursion error
            if not key.include and not key.anchor:
                prefix = list(
                    itertools.takewhile(lambda x: x[0] == x[1], zip(path, templatePath))
                )
                if len(prefix) == len(templatePath):
                    raise UnfurlError(
                        'recursive include "%s" in "%s" when including %s'
                        % (templatePath, path, key.key)
                    )
            includes = CommentedMap()
            template = expandDict(doc, path, includes, template, cls=cls)
    finally:
        if key.include:
            doc.loadTemplate(baseDir)  # pop baseDir
    return template


def _findTemplate(doc, key, path, cls, fail):
    template = doc
    templatePath = None
    if key.anchor:
        template = findAnchor(doc, key.anchor)
        if template is None:
            if not fail:
                return None
            else:
                raise UnfurlError("could not find anchor '%s'" % key.anchor)
        # XXX we don't know the path to the anchor so we can't support relative paths also
    elif key.relative:
        stop = None
        if key.relative - 1:
            stop = (key.relative - 1) * -1
        if templatePath is None:
            templatePath = list(path[:stop])
        else:
            templatePath = templatePath[:stop]
        template = lookupPath(doc, templatePath, cls)
        if template is None:
            if not fail:
                return None
            else:
                raise UnfurlError("could relative path '%s'" % ("." * key.relative))
    for index, segment in enumerate(key.pointer):
        if isinstance(template, Sequence):
            # Array indexes should be turned into integers
            try:
                segment = int(segment)
            except ValueError:
                pass
        try:
            template = template[segment]
        except (TypeError, LookupError):
            if not fail:
                return None
            raise UnfurlError(
                'can not find "%s" in document' % key.pointer[: index + 1].join("/")
            )

        if templatePath is not None:
            templatePath.append(segment)
    return template, templatePath


def hasTemplate(doc, key, path, cls):
    if key.include:
        return hasattr(doc, "loadTemplate")
    return _findTemplate(doc, key, path, cls, False) is not None


class _MissingInclude(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __repr__(self):
        return self.key.key


RE_FIRST = re.compile(r"([?]?)(include\d*)?([*]\S+)?([.]+$)?")

MergeKey = namedtuple("MergeKey", "key, maybe, include, anchor, relative, pointer")


def parseMergeKey(key):
    """
  +[maybe]?[include]?[anchor]?[relative]?[jsonpointer]?

  [include] = "include"[number?]
  [anchor] = "*"[anchorname]
  relative = '.'+
  """
    original = key
    key = key[1:]
    _jsonPointerValidate(key)
    parts = [_jsonPointerUnescape(part) for part in key.split("/")]
    first = parts.pop(0)
    maybe, include, anchor, relative = RE_FIRST.match(first).groups()
    if anchor:
        # relative will be included in anchor
        relative = len(anchor.rstrip(".")) - len(anchor)
        anchor = anchor[1:]  # exclude '*''
    elif relative:
        relative = len(relative)
    else:
        relative = 0
    if not (maybe or include or anchor or relative):
        if first:  # first wasn't empty and didn't match
            return None
        elif not parts:
            return None  # first was empty and pointer portion was too
    return MergeKey(original, not not maybe, include, anchor, relative, tuple(parts))


def expandDict(doc, path, includes, current, cls=dict):
    """
  Return a copy of `doc` that expands include directives.
  Include directives look like "+path/to/value"
  When appearing as a key in a map it will merge the result with the current dictionary.
  When appearing in a list it will insert the result in the list;
  if result is also a list, each item will be inserted separately.
  (If you don't want that behavior just wrap include in another list, e.g "[+list1]")
  """
    cp = cls()
    # first merge any includes includes into cp
    templates = []
    assert isinstance(current, Mapping), current
    for (key, value) in current.items():
        if not isinstance(key, six.string_types):
            cp[key] = value
            continue
        if key.startswith("+"):
            if key == mergeStrategyKey:
                cp[key] = value
                continue
            mergeKey = parseMergeKey(key)
            if not mergeKey:
                cp[key] = value
                continue
            foundTemplate = hasTemplate(doc, mergeKey, path, cls)
            if not foundTemplate:
                includes.setdefault(path, []).append(_MissingInclude(mergeKey, value))
                cp[key] = value
                continue
            includes.setdefault(path, []).append((mergeKey, value))
            template = getTemplate(doc, mergeKey, value, path, cls)
            if isinstance(template, Mapping):
                templates.append(template)
            else:
                if len(current) > 1:  # XXX include merge directive keys in count
                    raise UnfurlError("can not merge non-map value %s" % template)
                else:
                    return template  # current dict is replaced with a value
        # elif key.startswith("q+"):
        #    cp[key[2:]] = value
        elif isinstance(value, Mapping):
            cp[key] = expandDict(doc, path + (key,), includes, value, cls)
        elif isinstance(value, list):
            cp[key] = list(expandList(doc, path + (key,), includes, value, cls))
        else:
            cp[key] = value

    if templates:
        accum = templates.pop(0)
        templates.append(cp)
        while templates:
            cls = getattr(templates[0], "mapCtor", cls)
            accum = mergeDicts(accum, templates.pop(0), cls)
        return accum
    else:
        return cp
    # e,g, mergeDicts(mergeDicts(a, b), cp)
    # return includes, reduce(lambda accum, next: mergeDicts(accum, next, cls), templates, {}), cp


def _findMissingIncludes(includes):
    for x in includes.values():
        for i in x:
            if isinstance(i, _MissingInclude):
                yield i


def _deleteDeletedKeys(expanded):
    for key, value in expanded.items():
        if isinstance(value, Mapping):
            if value.get(mergeStrategyKey) == "delete":
                del expanded[key]
            else:
                _deleteDeletedKeys(value)


def expandDoc(doc, current=None, cls=dict):
    includes = CommentedMap()
    if current is None:
        current = doc
    if not isinstance(doc, Mapping) or not isinstance(current, Mapping):
        raise UnfurlError("top level element %s is not a dict" % doc)
    expanded = expandDict(doc, (), includes, current, cls)
    if hasattr(doc, "_anchorCache"):
        expanded._anchorCache = doc._anchorCache
    last = 0
    while True:
        missing = list(_findMissingIncludes(includes))
        if len(missing) == 0:
            # remove any stray keys with delete merge directive
            _deleteDeletedKeys(expanded)
            return includes, expanded
        if len(missing) == last:  # no progress
            raise UnfurlError("missing includes: %s" % missing)
        last = len(missing)
        includes = CommentedMap()
        expanded = expandDict(expanded, (), includes, current, cls)
        if hasattr(doc, "_anchorCache"):
            expanded._anchorCache = doc._anchorCache


def expandList(doc, path, includes, value, cls=dict):
    for i, item in enumerate(value):
        if isinstance(item, Mapping):
            if item.get(mergeStrategyKey) == "delete":
                continue
            newitem = expandDict(doc, path + (i,), includes, item, cls)
            if isinstance(newitem, MutableSequence):
                for i in newitem:
                    yield i
            else:
                yield newitem
        else:
            yield item


def diffDicts(old, new, cls=dict):
    """
  return a dict where old + diff = new
  """
    diff = cls()
    # start with old to preserve original order
    for key, val in old.items():
        if key in new:
            newval = new[key]
            if val != newval:
                if isinstance(val, Mapping) and isinstance(newval, Mapping):
                    diff[key] = diffDicts(val, newval, cls)
                else:
                    diff[key] = newval
        else:
            diff[key] = {"+%": "delete"}

    for key in new:
        if key not in old:
            diff[key] = new[key]
    return diff


# XXX rename function, confusing name
def patchDict(old, new, cls=dict):
    """
  Transform old into new while preserving old as much as possible.
  """
    # start with old to preserve original order
    for key, val in list(old.items()):
        if key in new:
            newval = new[key]
            if val != newval:
                if isinstance(val, Mapping) and isinstance(newval, Mapping):
                    old[key] = patchDict(val, newval, cls)
                elif isinstance(val, MutableSequence) and isinstance(
                    newval, MutableSequence
                ):
                    # preserve old item in list if they are equal to the new item
                    old[key] = [
                        (val[val.index(item)] if item in val else item)
                        for item in newval
                    ]
                else:
                    old[key] = newval
        else:
            del old[key]

    for key in new:
        if key not in old:
            old[key] = new[key]

    return old


def intersectDict(old, new, cls=dict):
    """
  remove keys from old that don't match new
  """
    # start with old to preserve original order
    for key, val in list(old.items()):
        if key in new:
            newval = new[key]
            if val != newval:
                if isinstance(val, Mapping) and isinstance(newval, Mapping):
                    old[key] = intersectDict(val, newval, cls)
                else:
                    del old[key]
        else:
            del old[key]

    return old


def lookupPath(doc, path, cls=dict):
    template = doc
    for segment in path:
        if isinstance(template, Sequence):
            try:
                segment = int(segment)
            except ValueError:
                return None
        try:
            template = template[segment]
        except (TypeError, LookupError):
            return None
    return template


def replacePath(doc, key, value, cls=dict):
    path = key[:-1]
    last = key[-1]
    ref = lookupPath(doc, path, cls)
    ref[last] = value


def addTemplate(changedDoc, path, mergeKey, template, cls):
    # if includeKey.anchor: #???
    if mergeKey.relative:
        if mergeKey.relative > 1:
            path = path[: (mergeKey.relative - 1) * -1]
        current = lookupPath(changedDoc, path, cls)
    else:
        current = changedDoc

    assert mergeKey.pointer
    path = mergeKey.pointer[:-1]
    last = mergeKey.pointer[-1]
    for segment in path:
        current = current.setdefault(segment, {})
    current[last] = template


def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
    """
  Modifies changedDoc with to use the includes found in originalDoc
  """
    # if the path to the include still exists
    # resolve the include
    # if the include doesn't exist in the current doc, re-add it
    # create a diff between the current object and the merged includes
    expandedOriginalIncludes, expandedOriginalDoc = expandDoc(originalDoc, cls=cls)
    for key, value in includes.items():
        ref = lookupPath(changedDoc, key, cls)
        if ref is None:
            # inclusion point no longer exists
            continue

        mergedIncludes = {}
        for (includeKey, includeValue) in value:
            if includeKey.include:
                ref = None
                continue
            stillHasTemplate = hasTemplate(changedDoc, includeKey, key, cls)
            if stillHasTemplate:
                template = getTemplate(changedDoc, includeKey, includeValue, key, cls)
            else:
                if hasTemplate(originalDoc, includeKey, key, cls):
                    template = getTemplate(
                        originalDoc, includeKey, includeValue, key, cls
                    )
                else:
                    template = getTemplate(
                        expandedOriginalDoc, includeKey, includeValue, key, cls
                    )

            if not isinstance(ref, Mapping):
                # XXX3 if isinstance(ref, list) lists not yet implemented
                if ref == template:
                    # ref still resolves to the template's value so replace it with the include
                    replacePath(changedDoc, key, {includeKey.key: includeValue}, cls)
                # ref isn't a map anymore so can't include a template
                break

            if not isinstance(template, Mapping):
                # ref no longer includes that template or we don't want to save it
                continue
            else:
                mergedIncludes = mergeDicts(mergedIncludes, template, cls)
                ref[includeKey.key] = includeValue

            if not stillHasTemplate:
                if includeValue != "raw":
                    if hasTemplate(originalDoc, includeKey, key, cls):
                        template = getTemplate(originalDoc, includeKey, "raw", key, cls)
                    else:
                        template = getTemplate(
                            expandedOriginalDoc, includeKey, "raw", key, cls
                        )
                addTemplate(changedDoc, key, includeKey, template, cls)

        if isinstance(ref, Mapping):
            diff = diffDicts(mergedIncludes, ref, cls)
            replacePath(changedDoc, key, diff, cls)



import unittest
import os
import unfurl.manifest
from unfurl.yamlmanifest import YamlManifest
from unfurl.localenv import LocalEnv
from unfurl.job import Runner, JobOptions
from unfurl.support import Status, _getbaseDir, RefContext
from unfurl.configurator import Configurator
from unfurl.util import sensitive_str, API_VERSION
from unfurl.yamlloader import makeVaultLib
import six
from click.testing import CliRunner

# python 2.7 needs these:
from unfurl.configurators.shell import ShellConfigurator


class SetAttributeConfigurator(Configurator):
    def run(self, task):
        from toscaparser.elements.portspectype import PortSpec

        if "ports" in task.inputs:
            ports = task.inputs["ports"]
            # target:source
            assert str(PortSpec(ports[0])) == "50000:9000", PortSpec(ports[0])
            assert str(PortSpec(ports[1])) == "20000-60000:1000-10000/udp", PortSpec(
                ports[1]
            )
            assert str(PortSpec(ports[2])) == "8000", PortSpec(ports[2])

        task.target.attributes["private_address"] = "10.0.0.1"
        yield task.done(True, Status.ok)


manifestDoc = """
apiVersion: unfurl/v1alpha1
kind: Manifest
context:
  inputs:
    cpus: 2
spec:
  service_template:
    node_types:
      testy.nodes.aNodeType:
        derived_from: tosca.nodes.Root
        requirements:
          - host:
              capabilities: tosca.capabilities.Compute
              relationship: tosca.relationships.HostedOn
        attributes:
          distribution:
            type: string
            default: { get_attribute: [ HOST, os, distribution ] }
        properties:
          private_address:
            type: string
            metadata:
              sensitive: true
          ports:
            type: list
            entry_schema:
              type: tosca.datatypes.network.PortSpec
          TEST_VAR:
            type: unfurl.datatypes.EnvVar
          vars:
            type: map
            entry_schema:
              type: unfurl.datatypes.EnvVar
          access_token:
            type: tosca.datatypes.Credential
          event_object: # 5.3.2.2 Examples p.194
            type: tosca.datatypes.json
            constraints:
              - schema: >
                  {
                    "$schema": "http://json-schema.org/draft-04/schema#",
                    "description": "Example Event type schema",
                    "type": "object",
                    "properties": {
                      "uuid": {
                        "description": "The unique ID for the event.",
                        "type": "string"
                      }
                    }
                  }
    topology_template:
      inputs:
        cpus:
          type: integer
          description: Number of CPUs for the server.
          constraints:
            - valid_values: [ 1, 2, 4, 8 ]
          metadata:
            sensitive: true
      outputs:
        server_ip:
          description: The private IP address of the provisioned server.
          value: { get_attribute: [ testSensitive, private_address ] }
          # equivalent to {eval: "::testSensitive::private_address"}

      node_templates:
        testSensitive:
          type: testy.nodes.aNodeType
          requirements:
            - host: my_server
          properties:
            private_address: foo
            ports:
              - source: 9000
                target: 50000
              - target_range: [ 20000, 60000 ]
                source_range: [ 1000, 10000 ]
                protocol: udp
              - source: 8000
            TEST_VAR: foo
            vars:
              VAR1: more
            access_token:
              protocol: xauth
              token_type: X-Auth-Token
              token: 604bbe45ac7143a79e14f3158df67091
            event_object: >
              {
                 "uuid": "cadf:1234-56-0000-abcd"
              }
          interfaces:
           Standard:
            create:
              inputs:
                ports: "{{ SELF['ports'] }}"
              implementation:
                primary: SetAttributeConfigurator
        my_server:
          type: tosca.nodes.Compute
          properties:
            test:  { concat: ['cpus: ', {get_input: cpus }] }
          capabilities:
            # Host container properties
            host:
             properties:
               num_cpus: { eval: ::inputs::cpus }
               disk_size: 10 GB
               mem_size: 512 MB
            # Guest Operating System properties
            os:
              properties:
                # host Operating System image properties
                architecture: x86_64
                type: Linux
                distribution: RHEL
                version: 6.5
          interfaces:
           Standard:
            create:
              implementation:
                primary: SetAttributeConfigurator
                timeout: 120
"""


class ToscaSyntaxTest(unittest.TestCase):
    def _runInputAndOutputs(self, manifest):
        job = Runner(manifest).run(JobOptions(add=True, startTime="time-to-test"))
        assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
        my_server = manifest.getRootResource().findResource("my_server")
        assert my_server
        assert my_server.attributes["test"], "cpus: 2"
        # print(job.out.getvalue())
        testSensitive = manifest.getRootResource().findResource("testSensitive")
        for name, toscaType in (
            ("access_token", "tosca.datatypes.Credential"),
            ("TEST_VAR", "unfurl.datatypes.EnvVar"),
        ):
            assert (
                testSensitive.template.attributeDefs[name].schema["type"] == toscaType
            )

        def t(datatype):
            return datatype.type == "unfurl.datatypes.EnvVar"

        envvars = set(testSensitive.template.findProps(testSensitive.attributes, t))
        self.assertEqual(envvars, set([("TEST_VAR", "foo"), ("VAR1", "more")]))
        outputIp = job.getOutputs()["server_ip"]
        self.assertEqual(outputIp, "10.0.0.1")
        assert isinstance(outputIp, sensitive_str), type(outputIp)
        assert job.status == Status.ok, job.summary()
        self.assertEqual("RHEL", testSensitive.attributes["distribution"])
        return outputIp, job

    def test_inputAndOutputs(self):
        manifest = YamlManifest(manifestDoc)
        outputIp, job = self._runInputAndOutputs(manifest)
        self.assertEqual(
            ["my_server", "testSensitive"],
            job.rootResource.query(
                "$nodes::.name",
                vars=dict(nodes={"get_nodes_of_type": "tosca.nodes.Root"}),
            ),
        )
        self.assertEqual(
            "testSensitive",
            job.rootResource.query({"get_nodes_of_type": "testy.nodes.aNodeType"})[
                0
            ].name,
        )
        assert not job.rootResource.query(
            {"get_nodes_of_type": "testy.nodes.Nonexistent"}
        )
        assert "server_ip: <<REDACTED>>" in job.out.getvalue(), job.out.getvalue()

    def test_ansibleVault(self):
        manifest = YamlManifest(manifestDoc, vault=makeVaultLib("a_password"))
        outputIp, job = self._runInputAndOutputs(manifest)
        vaultString = "server_ip: !vault |\n      $ANSIBLE_VAULT;1.1;AES256"
        assert vaultString in job.out.getvalue(), job.out.getvalue()

    def test_import(self):
        """
      Tests nested imports and url fragment resolution.
      """
        path = __file__ + "/../examples/testimport-manifest.yaml"
        manifest = YamlManifest(path=path)
        self.assertEqual(2, len(manifest.tosca.template.nested_tosca_tpls.keys()))

        runner = Runner(manifest)
        output = six.StringIO()
        job = runner.run(JobOptions(add=True, out=output, startTime="test"))
        self.assertEqual(job.status.name, "ok")
        self.assertEqual(job.stats()["ok"], 1)
        self.assertEqual(job.getOutputs()["aOutput"], "set")
        assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
        # print(output.getvalue())
        anInstance = job.rootResource.findResource("testPrefix")
        assert anInstance
        ctx = RefContext(anInstance)

        # .: <ensemble>/
        base = _getbaseDir(ctx, ".")
        self.assertEqual(base, os.path.normpath(os.path.dirname(path)))

        # testPrefix appeared in the same source file so it will be the same
        src = _getbaseDir(ctx, "src")
        self.assertEqual(src, base)

        # home: <ensemble>/<instance name>/
        home = _getbaseDir(ctx, "home")
        self.assertEqual(os.path.join(base, "testPrefix"), home)

        # local: <ensemble>/<instance name>/local/
        local = _getbaseDir(ctx, "local")
        self.assertEqual(os.path.join(home, "local"), local)

        tmp = _getbaseDir(ctx, "tmp")
        assert tmp.endswith("testPrefix"), tmp

        # spec.home: <spec>/<template name>/
        specHome = _getbaseDir(ctx, "spec.home")
        self.assertEqual(os.path.join(base, "spec", "testPrefix"), specHome)

        # spec.local: <spec>/<template name>/local/
        specLocal = _getbaseDir(ctx, "spec.local")
        self.assertEqual(os.path.join(specHome, "local"), specLocal)

        specSrc = _getbaseDir(ctx, "spec.src")
        self.assertEqual(src, specSrc)

        # these repositories should always be defined:
        unfurlRepoPath = _getbaseDir(ctx, "unfurl")
        self.assertEqual(unfurl.manifest._basepath, os.path.normpath(unfurlRepoPath))

        spec = _getbaseDir(ctx, "spec")
        self.assertEqual(os.path.normpath(spec), base)

        selfPath = _getbaseDir(ctx, "self")
        self.assertEqual(os.path.normpath(selfPath), base)

    def test_workflows(self):
        manifest = YamlManifest(
            path=__file__ + "/../examples/test-workflow-manifest.yaml"
        )
        # print(manifest.tosca.template.nested_tosca_tpls)
        self.assertEqual(len(manifest.tosca._workflows), 3)

        runner = Runner(manifest)
        output = six.StringIO()
        job = runner.run(
            JobOptions(add=True, planOnly=True, out=output, startTime="test")
        )
        # print(job.jsonSummary())
        assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
        self.assertEqual(job.status.name, "ok")
        self.assertEqual(job.stats()["ok"], 4)
        self.assertEqual(job.stats()["changed"], 4)


class AbstractTemplateTest(unittest.TestCase):
    def test_import(self):
        foreign = (
            """
    apiVersion: %s
    kind: Manifest
    spec:
      service_template:
        node_types:
          test.nodes.AbstractTest:
            derived_from: tosca.nodes.Root
            properties:
              private_address:
                type: string
                required: false
                metadata:
                  sensitive: true
            interfaces:
               Install:
                operations:
                 check:
                   implementation: SetAttributeConfigurator
      instances:
        anInstance:
         type: test.nodes.AbstractTest
    """
            % API_VERSION
        )

        localConfig = """
          apiVersion: unfurl/v1alpha1
          kind: Project
          contexts:
            defaults:
              secrets:
                attributes:
                  vault_default_password: a_password
              external:
               foreign:
                  manifest:
                    file:  foreignmanifest.yaml
                  instance: "*"  # this is the default
        """

        # import a node from a external manifest and have an abstract node template select it
        # check will be run on it each time
        mainManifest = (
            """
apiVersion: %s
kind: Manifest
spec:
  service_template:
    imports:
      - foreignmanifest.yaml#/spec/service_template
    topology_template:
      outputs:
        server_ip:
          value: {eval: "::foreign:anInstance::private_address"}
      node_templates:
        anInstance:
          type: test.nodes.AbstractTest
          directives:
             - select
  """
            % API_VERSION
        )

        runner = CliRunner()  # delete UNFURL_HOME
        try:
            UNFURL_HOME = os.environ.get("UNFURL_HOME")
            with runner.isolated_filesystem():
                os.environ["UNFURL_HOME"] = ""

                with open("foreignmanifest.yaml", "w") as f:
                    f.write(foreign)

                with open("unfurl.yaml", "w") as f:
                    f.write(localConfig)

                with open("manifest.yaml", "w") as f:
                    f.write(mainManifest)

                manifest = LocalEnv("manifest.yaml").getManifest()
                assert manifest.manifest.vault and manifest.manifest.vault.secrets
                job = Runner(manifest).run(
                    JobOptions(add=True, startTime="time-to-test")
                )
                # print(job.out.getvalue())
                # print(job.jsonSummary(True))
                assert job.status == Status.ok, job.summary()
                self.assertEqual(
                    [
                        {
                            "operation": "check",
                            "configurator": "SetAttributeConfigurator",
                            "changed": True,
                            "priority": "required",
                            "reason": "check",
                            "status": "ok",
                            "target": "foreign:anInstance",
                            "targetStatus": "ok",
                            "template": "anInstance",
                            "type": "test.nodes.AbstractTest",
                        }
                    ],
                    job.jsonSummary()["tasks"],
                )
                job.getOutputs()
                self.assertEqual(job.getOutputs()["server_ip"], "10.0.0.1")
                self.assertEqual(
                    len(manifest.localEnv._manifests), 2, manifest.localEnv._manifests
                )
                # print("output", job.out.getvalue())
                assert "10.0.0.1" not in job.out.getvalue(), job.out.getvalue()
                vaultString1 = "server_ip: !vault |\n      $ANSIBLE_VAULT;1.1;AES256"
                assert vaultString1 in job.out.getvalue()
                vaultString2 = (
                    "private_address: !vault |\n          $ANSIBLE_VAULT;1.1;AES256"
                )
                assert vaultString2 in job.out.getvalue()

                # reload:
                manifest2 = LocalEnv("manifest.yaml").getManifest()
                assert manifest2.lastJob
                # test that restored manifest create a shadow instance for the foreign instance
                imported = manifest2.imports["foreign"].resource
                assert imported
                imported2 = manifest2.imports.findImport("foreign:anInstance")
                assert imported2
                assert imported2.shadow
                self.assertIs(imported2.root, manifest2.getRootResource())
                self.assertEqual(imported2.attributes["private_address"], "10.0.0.1")
                self.assertIsNot(imported2.shadow.root, manifest2.getRootResource())
        finally:
            if UNFURL_HOME is not None:
                os.environ["UNFURL_HOME"] = UNFURL_HOME

    def test_connections(self):
        mainManifest = (
            """
apiVersion: %s
kind: Manifest
spec:
  service_template:
    topology_template:
      node_templates:
        defaultCluster:
          directives:
            - default
          type: unfurl.nodes.K8sCluster
        myCluster:
          type: unfurl.nodes.K8sCluster
        localhost:
          type: unfurl.nodes.Default
          requirements:
            - connect: # section 3.8.2 p140
                relationship:
                  # think of this as a connection, "discover" will figure out what's on the other end
                  type: unfurl.relationships.ConnectsTo.K8sCluster
                  properties:
                    context: docker-desktop
  """
            % API_VERSION
        )
        manifest2 = YamlManifest(mainManifest)
        nodeSpec = manifest2.tosca.getTemplate("localhost")
        assert nodeSpec
        relationshipSpec = nodeSpec.requirements["connect"].relationship
        assert relationshipSpec
        self.assertEqual(relationshipSpec.name, "connect")
        self.assertEqual(
            relationshipSpec.type, "unfurl.relationships.ConnectsTo.K8sCluster"
        )
        # chooses myCluster instead of the cluster with the "default" directive
        assert relationshipSpec.target.name == "myCluster"
